#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"
#include "xf86Pci.h"
#include "xf86PciInfo.h"
#include "xf86Xinput.h"
#include "xf86fbman.h"
#include "xf86i2c.h"
#include "xf86xv.h"
#include "Xv.h"

#include "atiregs.h"
#include "ati.h"
#include "atiident.h"
#include "atipriv.h"
#include "atistruct.h"
#include "atichip.h"
#include "atimach64.h"
#include "atiio.h"

/*
 * Note:  Only 32-bit MMIO is needed here.
 */

#define inm(_Register)                                        \
    MMIO_IN32(pATI->pBlock[GetBits(_Register, BLOCK_SELECT)], \
              (_Register) & MM_IO_SELECT)


#define CacheByte(___Register) pATI->MMIOCached[CacheSlotOf(___Register) >> 3]
#define CacheBit(___Register)  (0x80U >> (CacheSlotOf(___Register) & 0x07U))

#define RegisterIsCached(__Register) \
    (CacheByte(__Register) & CacheBit(__Register))
#define CacheRegister(__Register) \
    CacheByte(__Register) |= CacheBit(__Register)
#define UnCacheRegister(__Register) \
    CacheByte(__Register) &= ~CacheBit(__Register)

#define CacheSlot(__Register)  pATI->MMIOCache[CacheSlotOf(__Register)]

#define ATIMach64WaitForFIFO(_n)                               \
    while ((pATI->nAvailableFIFOEntries < (_n)) &&             \
           (pATI->nAvailableFIFOEntries < pATI->nFIFOEntries)) \
        ATIMach64PollEngineStatus(pATI)

#define ATIMach64WaitForIdle()                 \
    while (pATI->EngineIsBusy)                 \
        ATIMach64PollEngineStatus(pATI)



/* This would be quite a bit slower as a function */
#define outm(_Register, _Value)                                        \
    do                                                                 \
    {                                                                  \
        CARD32 _IOValue = (_Value);                                    \
                                                                       \
        if (!RegisterIsCached(_Register) ||                            \
            (_IOValue != CacheSlot(_Register)))                        \
        {                                                              \
            while (!pATI->nAvailableFIFOEntries--)                     \
                ATIMach64PollEngineStatus(pATI);                       \
            MMIO_OUT32(pATI->pBlock[GetBits(_Register, BLOCK_SELECT)], \
                (_Register) & MM_IO_SELECT, _IOValue);                 \
            CacheSlot(_Register) = _IOValue;                           \
            pATI->EngineIsBusy = TRUE;                                 \
        }                                                              \
    } while (0)



#define XV_ENCODING	"XV_ENCODING"
#define XV_BRIGHTNESS  	"XV_BRIGHTNESS"
#define XV_CONTRAST 	"XV_CONTRAST"
#define XV_SATURATION  	"XV_SATURATION"
#define XV_HUE		"XV_HUE"
#define XV_FORMAT		"XV_FORMAT"
#define XV_COLOR_KEY		"XV_COLOR_KEY"

#define MAKE_ATOM(a) MakeAtom(a, sizeof(a) - 1, TRUE)

static Atom xvEncoding, xvBrightness, xvContrast, xvSaturation, xvHue;
static Atom xv_format, xv_color_key;

#define NUM_ATI_ATTRIBUTES_PORT0	2

XF86AttributeRec ati_attributes_port0[NUM_ATI_ATTRIBUTES_PORT0]={
	{XvSettable,0,-1,XV_FORMAT},
	{XvSettable,0,-1,XV_COLOR_KEY}
	};

typedef struct _PortPrivRec {
    struct _AdaptorPrivRec *    pAdaptor;
/*    I2CDevRec                   I2CDev;
    int                         Plug;
    INT32			Attribute[8];		/* Brig, Con, Sat, Hue, Int, Filt */
    FBAreaPtr			pFBArea[2];
    int				Buffers;


    long			buf_offset[2];
    
    CARD16			vert_inc;
    CARD16			horz_inc;

    CARD32			overlay_scale_cntl;
    CARD32			video_format;
    CARD32			capture_config;
    CARD32			overlay_key;
    CARD8			overlay_pixel_size;
    CARD8			current_buffer;
    int				overlay_pad;
    CARD32			overlay_id;
    CARD32			overlay_width;
    CARD32			overlay_height;
    CARD32			BufferBase[3];		/* x1 */
    CARD32			BufferStride;		/* x1 */

    CARD32			drw_w,drw_h;
    
    CARD32			save_off_pitch,save_h_total_disp,save_v_total_disp;

   /* CARD32			BufferPProd;		/* PProd(BufferStride / 2) */
    INT32			vx, vy, vw, vh;		/* 12.10 fp; int */
    int				dx, dy, dw, dh;
    int				fw, fh;
/*    CookiePtr			pCookies; 
    int				nCookies;
    INT32			dS, dT;			/* 12.20 fp */
    int				APO;
    int				FramesPerSec, FrameAcc;
    int				BkgCol;			/* RGB 5:6:5;5:6:5 */
    int				VideoOn;		/* No, Once, Yes */
    Bool			StreamOn;
} PortPrivRec, *PortPrivPtr;


#define HW_UNSUPPORTED	0
#define HW_MACH64	1
#define HW_RAGEPRO	2


typedef struct _AdaptorPrivRec {
    struct _AdaptorPrivRec *	Next;
    ScrnInfoPtr			pScrn;
/*    void *			pm2p;
    LFBAreaPtr			LFBList;
    CARD32			FifoControl;
    OsTimerPtr			Timer;
    int                         VideoStd;
    int				FramesPerSec;
    int				FrameLines;
    int				IntLine;		/* Frame */
/*    int				LinePer;		/* nsec */
    int				hw_type;  /* hardware type */
    PortPrivRec                 Port[2]; 
} AdaptorPrivRec, *AdaptorPrivPtr;


static XF86VideoEncodingRec
InputVideoEncodings[] =
{
    { 0, "pal-composite",		704, 576, { 1, 50 }},
    { 1, "pal-composite_adaptor",	704, 576, { 1, 50 }},
    { 2, "pal-svideo",			704, 576, { 1, 50 }},
    { 3, "ntsc-composite",		704, 480, { 1001, 60000 }},
    { 4, "ntsc-composite_adaptor",	704, 480, { 1001, 60000 }},
    { 5, "ntsc-svideo",			704, 480, { 1001, 60000 }},
    { 6, "secam-composite",		704, 576, { 1, 50 }},
    { 7, "secam-composite_adaptor",	704, 576, { 1, 50 }},
    { 8, "secam-svideo",		704, 576, { 1, 50 }},
    { 9,  "XV_IMAGE",			-1,-1,{-1,-1}}
};

static XF86VideoEncodingRec
OutputVideoEncodings[] =
{
    { 0, "pal-composite_adaptor",	704, 576, { 1, 50 }},
    { 1, "pal-svideo",			704, 576, { 1, 50 }},
    { 2, "ntsc-composite_adaptor",	704, 480, { 1001, 60000 }},
    { 3, "ntsc-svideo",			704, 480, { 1001, 60000 }},
};

static XF86VideoFormatRec
InputVideoFormats[] = {
    { 8,  TrueColor }, /* Dithered */
    { 15, TrueColor },
    { 16, TrueColor },
    { 24, TrueColor },
};

static XF86VideoFormatRec
OutputVideoFormats[] = {
    { 8,  TrueColor },
    { 8,  PseudoColor }, /* Using .. */
    { 8,  StaticColor },
    { 8,  GrayScale },
    { 8,  StaticGray }, /* .. TexelLUT */
    { 15, TrueColor },
    { 16, TrueColor },
    { 24, TrueColor },
};

#define NUM_IMAGES_G 3

static XF86ImageRec ImagesG[NUM_IMAGES_G] =
{
   {
	0x32595559,
        XvYUV,
	LSBFirst,
	{'Y','U','Y','2',
	  0x00,0x00,0x00,0x10,0x80,0x00,0x00,0xAA,0x00,0x38,0x9B,0x71},
	16,
	XvPacked,
	1,
	0, 0, 0, 0 ,
	8, 8, 8, 
	1, 2, 2,
	1, 1, 1,
	{'Y','U','Y','V',
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	XvTopToBottom
   },
   {
	0x32315659,
        XvYUV,
	LSBFirst,
	{'Y','V','1','2',
	  0x00,0x00,0x00,0x10,0x80,0x00,0x00,0xAA,0x00,0x38,0x9B,0x71},
	12,
	XvPlanar,
	3,
	0, 0, 0, 0 ,
	8, 8, 8, 
	1, 2, 2,
	1, 2, 2,
	{'Y','V','U',
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	XvTopToBottom
   },
   {
	0x59565955,
        XvYUV,
	LSBFirst,
	{'U','Y','V','Y',
	  0x00,0x00,0x00,0x10,0x80,0x00,0x00,0xAA,0x00,0x38,0x9B,0x71},
	16,
	XvPacked,
	1,
	0, 0, 0, 0 ,
	8, 8, 8, 
	1, 2, 2,
	1, 1, 1,
	{'U','Y','V','Y',
	  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	XvTopToBottom
   }
};


static void FreeOffscreenBuffers(PortPrivPtr pPPriv)
{
AdaptorPrivPtr pAPriv=pPPriv->pAdaptor;
ScrnInfoPtr pScrn=pAPriv->pScrn;
int buffer;
for(buffer=0;buffer<pPPriv->Buffers;buffer++)
if(pPPriv->pFBArea[buffer]!=NULL){
	xf86FreeOffscreenArea(pPPriv->pFBArea[buffer]);
	pPPriv->pFBArea[buffer]=NULL;
	}
}

static Bool AllocateOffscreenBuffers(PortPrivPtr pPPriv)
{
AdaptorPrivPtr pAPriv=pPPriv->pAdaptor;
ScrnInfoPtr pScrn=pAPriv->pScrn;
long length;
int buffer;
ATIPtr	pATI=ATIPTR(pScrn);
FreeOffscreenBuffers(pPPriv);
length=(pPPriv->overlay_pixel_size*pPPriv->overlay_width+pPPriv->overlay_pad)*
			pPPriv->overlay_height;
pPPriv->Buffers=2;			
for(buffer=0;buffer<pPPriv->Buffers;buffer++){
	pPPriv->pFBArea[buffer]=xf86AllocateLinearOffscreenArea(pScrn->pScreen,length,1,NULL,NULL,NULL);
	if(pPPriv->pFBArea[buffer]==NULL){
		xf86DrvMsg(pScrn->scrnIndex,X_ERROR,"Could not allocate buffer (%d bytes)\n",length);
		return FALSE;
		}
	pPPriv->buf_offset[buffer]=(pPPriv->pFBArea[buffer]->box.y1 *pScrn->displayWidth+
			pPPriv->pFBArea[buffer]->box.x1);
	switch(pScrn->depth){
		case 8:
			break;
		case 15:
		case 16:
			pPPriv->buf_offset[buffer]*=2;
			break;
		case 24:
			pPPriv->buf_offset[buffer]*=3;
			break;
		case 32:
			pPPriv->buf_offset[buffer]*=4;
			break;
		default:
		}
	pPPriv->buf_offset[buffer]&=~7;  /* debug */
	/* pPPriv->buf_offset[buffer]=0; */ /* debug */
	xf86DrvMsg(pScrn->scrnIndex,X_INFO,"Allocated buffer=%d offset=0x%08x\n",buffer,pPPriv->buf_offset[buffer]);
	}
outm(SCALER_BUF0_OFFSET,pPPriv->buf_offset[0]);
outm(SCALER_BUF0_OFFSET_U,pPPriv->buf_offset[0]);
outm(SCALER_BUF0_OFFSET_V,pPPriv->buf_offset[0]);

outm(SCALER_BUF1_OFFSET,pPPriv->buf_offset[1]);
outm(SCALER_BUF1_OFFSET_U,pPPriv->buf_offset[1]);
outm(SCALER_BUF1_OFFSET_V,pPPriv->buf_offset[1]);
return TRUE;
}

#define AtiEnableBits	((1<<31)|(1<<30))

#define AtiOverlayOn()        { \
			pPPriv->overlay_scale_cntl|=AtiEnableBits;\
			outm(OVERLAY_SCALE_CNTL,pPPriv->overlay_scale_cntl);\
			}


#define AtiOverlayOff()	{\
			pPPriv->overlay_scale_cntl&=~AtiEnableBits;\
			outm(OVERLAY_SCALE_CNTL,pPPriv->overlay_scale_cntl);\
			}

static int
AtiPutVideo(ScrnInfoPtr pScrn,
    short vid_x, short vid_y, short drw_x, short drw_y,
    short vid_w, short vid_h, short drw_w, short drw_h,
    RegionPtr clipBoxes, pointer data)
{
PortPrivPtr pPPriv = (PortPrivPtr) data;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ATIPtr	pATI=ATIPTR(pScrn);
int i;

xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiPutVideo\n");
return BadValue;
}

static int
AtiPutStill(ScrnInfoPtr pScrn,
    short vid_x, short vid_y, short drw_x, short drw_y,
    short vid_w, short vid_h, short drw_w, short drw_h,
    RegionPtr clipBoxes, pointer data)
{
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiPutStill\n");

return Success;
}

static int
AtiGetVideo(ScrnInfoPtr pScrn,
    short vid_x, short vid_y, short drw_x, short drw_y,
    short vid_w, short vid_h, short drw_w, short drw_h,
    RegionPtr clipBoxes, pointer data)
{
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiGetVideo\n");

return BadValue;
}

static int
AtiGetStill(ScrnInfoPtr pScrn,
    short vid_x, short vid_y, short drw_x, short drw_y,
    short vid_w, short vid_h, short drw_w, short drw_h,
    RegionPtr clipBoxes, pointer data)
{
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiGetStill\n");
return BadMatch;
}

static void
AtiStopVideo(ScrnInfoPtr pScrn, pointer data, Bool exit)
{
PortPrivPtr pPPriv = (PortPrivPtr) data;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ATIPtr	pATI=ATIPTR(pScrn);
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiStopVideo\n");
AtiOverlayOff();
if(exit)FreeOffscreenBuffers(pPPriv);
}

static int
AtiSetPortAttribute(ScrnInfoPtr pScrn,
    Atom attribute, INT32 value, pointer data)
{
PortPrivPtr pPPriv = (PortPrivPtr) data;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ATIPtr	pATI=ATIPTR(pScrn);
if(attribute==xv_color_key){
	xf86DrvMsg(pScrn->scrnIndex,X_INFO,"AtiSetPortAttribute color key 0x%08x\n",value);
	pPPriv->overlay_key=value;
	outm(OVERLAY_GRAPHICS_KEY_CLR,pPPriv->overlay_key);
	return Success;
	} 
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiSetPortAttribute unknown\n");
return BadMatch;
}

static int
AtiGetPortAttribute(ScrnInfoPtr pScrn, 
    Atom attribute, INT32 *value, pointer data)
{
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiGetPortAttribute\n");

return BadMatch;
}

static void
AtiQueryBestSize(ScrnInfoPtr pScrn, Bool motion,
    short vid_w, short vid_h, short drw_w, short drw_h,
    unsigned int *p_w, unsigned int *p_h, pointer data)
{
*p_w=drw_w;
*p_h=drw_h;
}

static void
AtiCopyMungedData(
   unsigned char *src1,
   unsigned char *src2,
   unsigned char *src3,
   unsigned char *dst1,
   int srcPitch,
   int srcPitch2,
   int dstPitch,
   int h,
   int w
){
   CARD32 *dst = (CARD32*)dst1;
   int i, j;

   dstPitch >>= 2;
   w >>= 1;

   for(j = 0; j < h; j++) {
	for(i = 0; i < w; i++) {
	    dst[i] = src1[i << 1] | (src1[(i << 1) + 1] << 16) |
		     (src3[i] << 8) | (src2[i] << 24);
	}
	dst += dstPitch;
	src1 += srcPitch;
	if(j & 1) {
	    src2 += srcPitch2;
	    src3 += srcPitch2;
	}
   }
}


static void AtiCopyData(
  unsigned char *src,
  unsigned char *dst,
  int srcPitch,
  int dstPitch,
  int h,
  int w
){
    w <<= 1;
    while(h--) {
	memcpy(dst, src, w);
	src += srcPitch;
	dst += dstPitch;
    }
}


int AtiPutImage( ScrnInfoPtr pScrn, 
	short src_x, short src_y, short drw_x, short drw_y,
	short src_w, short src_h, short drw_w, short drw_h,
	int id, unsigned char* buf, short width, short height, Bool sync,
	RegionPtr clipBoxes, pointer data )
{
PortPrivPtr pPPriv = (PortPrivPtr) data;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ATIPtr	pATI=ATIPTR(pScrn);
int i;
int srcPitch,srcPitch2;
int dstPitch;
int pitch,new_h,offset,offset2,offset3;
CARD32 tmp;
int top,left,npixels,nlines;
INT32 x1,x2,y1,y2;
CARD32 video_format;

/*
xf86DrvMsg(pScrn->scrnIndex,X_INFO,"AtiPutImage src_x=%d src_w=%d width=%d height=%d sync=%d\n",src_x,src_w,width,height,sync);
*/

switch(id) {
   	case 0x32315659:
		video_format=0xB000B;
		break;
	case 0x59565955:
		video_format=0xB000B;
		break;
   	case 0x32595559:
		video_format=0xC000C;
		break;
   	default:
		return BadValue;
  		
   	} 
if(video_format!=pPPriv->video_format){
	xf86DrvMsg(pScrn->scrnIndex,X_INFO,"AtiPutImage new format: id=0x%08x format=0x%08x\n",id,video_format);
	}


if((pPPriv == & pAPriv->Port[0])){
	if((width!=pPPriv->overlay_width)||
	   (height!=pPPriv->overlay_height)||
	   (id!=pPPriv->overlay_id)||
	   (video_format!=pPPriv->video_format)||
	   (pPPriv->overlay_pad<0)){
		FreeOffscreenBuffers(pPPriv);
		pPPriv->overlay_pad=0;
		pPPriv->overlay_width=width;
		pPPriv->overlay_height=height;
		pPPriv->overlay_pixel_size=2;
		pPPriv->overlay_id=id;
		AllocateOffscreenBuffers(pPPriv);
		}
   	x1 = src_x;
   	x2 = src_x + src_w;
   	y1 = src_y;
   	y2 = src_y + src_h;
	
	
	
	dstPitch=pPPriv->overlay_width*pPPriv->overlay_pixel_size;
	srcPitch=width;
    	/* copy data */
    	top = y1 >> 16;
    	left = (x1 >> 16) & ~1;
    	npixels = ((((x2 + 0xffff) >> 16) + 1) & ~1) - left;
    	left <<= 1;
	switch(id) {
	   	case 0x32315659:
			srcPitch = (width + 3) & ~3;
			offset2 = srcPitch * height;
			srcPitch2 = ((width >> 1) + 3) & ~3;
			offset3 = (srcPitch2 * (height >> 1)) + offset2;
			nlines = ((((y2 + 0xffff) >> 16) + 1) & ~1) - top;			
			break;
	   	case 0x59565955:
	   	case 0x32595559:
	   	default:
			buf += (top * srcPitch) + left;
			nlines = ((y2 + 0xffff) >> 16) - top;
			srcPitch = (width << 1);
    			break;
   		} 
	if(video_format!=pPPriv->video_format){
		pPPriv->video_format=video_format;
		outm(VIDEO_FORMAT,pPPriv->video_format);
		}
	nlines=src_h; /* ?*/
	npixels=src_w;
	if(npixels>pPPriv->overlay_width)npixels=pPPriv->overlay_width;
	if(nlines>pPPriv->overlay_height)nlines=pPPriv->overlay_height;
	
	pPPriv->vert_inc=(pPPriv->overlay_height<<12)/drw_h;
	pPPriv->horz_inc=(pPPriv->overlay_width<<12)/drw_w;
	pPPriv->drw_w=drw_w;
	pPPriv->drw_h=drw_h;
		
	if(pPPriv->current_buffer==(inm(CAPTURE_CONFIG) &(1<<29))){
		return Success; /* we are too fast */
		}

	outm(OVERLAY_Y_X_START,((drw_x)<<16)|(drw_y)|(1<<31));
	outm(OVERLAY_Y_X_END,((drw_x+drw_w)<<16)|(drw_y+drw_h));
	outm(OVERLAY_SCALE_INC,(pPPriv->horz_inc<<16)|(pPPriv->vert_inc));
	outm(SCALER_HEIGHT_WIDTH,(pPPriv->overlay_width<<16)|(pPPriv->overlay_height));
	outm(SCALER_BUF_PITCH,pPPriv->overlay_width);
	
	
	switch(id) {
	   	case 0x32315659:
			AtiCopyMungedData(buf + (top * srcPitch) + (left >> 1), 
				  buf + offset2, buf + offset3, pATI->pMemory+pPPriv->buf_offset[pPPriv->current_buffer],
				  srcPitch, srcPitch2, dstPitch, nlines, npixels);
			break;
		case 0x59565955:
	   	case 0x32595559:
	   	default:
			AtiCopyData(buf,pATI->pMemory+pPPriv->buf_offset[pPPriv->current_buffer],srcPitch,dstPitch,nlines,npixels);
  		
   	} 
	outm(CAPTURE_CONFIG, pPPriv->capture_config | (pPPriv->current_buffer<<29));
	pPPriv->current_buffer=!pPPriv->current_buffer;
	
	AtiOverlayOn();
	return Success;
	}
return BadMatch;
}

int AtiReputImage( ScrnInfoPtr pScrn, short drw_x, short drw_y,
        RegionPtr clipBoxes, pointer data )
{
PortPrivPtr pPPriv = (PortPrivPtr) data;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ATIPtr	pATI=ATIPTR(pScrn);
if(pPPriv == & pAPriv->Port[0]){
	outm(OVERLAY_Y_X_START,((drw_x)<<16)|(drw_y)|(1<<31));
	outm(OVERLAY_Y_X_END,((drw_x+pPPriv->drw_w)<<16)|(drw_y+pPPriv->drw_h));
	return Success;
	}
return BadMatch;
}

static int 
AtiQueryImageAttributesG(ScrnInfoPtr pScrn, 
  int id, 
  unsigned short *w, unsigned short *h, 
  int *pitches, int *offsets)
{
    int size, tmp;

    if(*w > 1024) *w = 1024;
    if(*h > 1024) *h = 1024;

    *w = (*w + 1) & ~1;
    if(offsets) offsets[0] = 0;

    switch(id) {
    case 0x32315659:
	*h = (*h + 1) & ~1;
	size = (*w + 3) & ~3;
	if(pitches) pitches[0] = size;
	size *= *h;
	if(offsets) offsets[1] = size;
	tmp = ((*w >> 1) + 3) & ~3;
	if(pitches) pitches[1] = pitches[2] = tmp;
	tmp *= (*h >> 1);
	size += tmp;
	if(offsets) offsets[2] = size;
	size += tmp;
	break;
    case 0x59565955:
    case 0x32595559:
    default:
	size = *w << 1;
	if(pitches) pitches[0] = size;
	size *= *h;
	break;
    }

    return size;
}


static void
AdaptorPrivUninit(AdaptorPrivPtr pAPriv)
{
/*
if(!pAPriv->atip){
	xf86DestroyI2CDev(&pAPriv->Port[0].I2CDev,FALSE);
	}
*/
xf86DrvMsgVerb(pAPriv->pScrn->scrnIndex,X_INFO,1,"ati_video: AdaptorPrivUninit\n");

xfree(pAPriv);
}

static int m64_pollscreen(ScrnInfoPtr pScrn,AdaptorPrivPtr pAPriv,int sig)
{
ATIPtr pATI=ATIPTR(pScrn);
CARD8 save,pll_vclk_cntl, pll_ref_div, vclk_post_div, vclk_fb_div,ecp_div;
int i;
double dotclock,refclock;
#define port   (pAPriv->Port[0])

if(sig  && (inm(pATI->CPIO_CRTC_OFF_PITCH)==port.save_off_pitch) 
	&& (inm(pATI->CPIO_CRTC_H_TOTAL_DISP==port.save_h_total_disp))
	&& (inm(pATI->CPIO_CRTC_V_TOTAL_DISP==port.save_v_total_disp))) return 0;

port.save_off_pitch=inm(pATI->CPIO_CRTC_OFF_PITCH);
port.save_h_total_disp=inm(pATI->CPIO_CRTC_H_TOTAL_DISP);
port.save_v_total_disp=inm(pATI->CPIO_CRTC_V_TOTAL_DISP);

save=inb(pATI->CPIO_CLOCK_CNTL+1);
i=inb(pATI->CPIO_CLOCK_CNTL) & 3;

outb(pATI->CPIO_CLOCK_CNTL+1,2<<2); pll_ref_div=inb(pATI->CPIO_CLOCK_CNTL+2);
outb(pATI->CPIO_CLOCK_CNTL+1,5<<2); pll_vclk_cntl=inb(pATI->CPIO_CLOCK_CNTL+2);
outb(pATI->CPIO_CLOCK_CNTL+1,6<<2); vclk_post_div=(inb(pATI->CPIO_CLOCK_CNTL+2)>>(i+i))&3;
outb(pATI->CPIO_CLOCK_CNTL+1,(7+i)<<2); vclk_fb_div=inb(pATI->CPIO_CLOCK_CNTL+2);
refclock=(pATI->ReferenceNumerator*1.0)/(157500.0/14.31818);
dotclock=2.0*refclock*vclk_fb_div / (pll_ref_div*(1<<vclk_post_div));
ecp_div=dotclock/(125.0);
xf86DrvMsg(pScrn->scrnIndex,X_INFO,"m64_pollscreen dotclock=%g ecp_div=%d\n",dotclock,ecp_div);
if(ecp_div>2)ecp_div=2;
if((pll_vclk_cntl & 0x03)!=0x03)ecp_div=0;
if((pll_vclk_cntl & 0x30)!=ecp_div<<4){
	outb(pATI->CPIO_CLOCK_CNTL+1,(5<<2)|2);
	outb(pATI->CPIO_CLOCK_CNTL+2,(pll_vclk_cntl & 0xCF) | (ecp_div<<4));
	}
	
outb(pATI->CPIO_CLOCK_CNTL+1,save);
#undef port
}

static AdaptorPrivPtr
AdaptorPrivInit(ScrnInfoPtr pScrn)
{
ATIPtr	pATI=ATIPTR(pScrn);
AdaptorPrivPtr pAPriv;

pAPriv=(AdaptorPrivPtr) xcalloc(1,sizeof(AdaptorPrivRec));
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"ati_video: AdaptorPrivInit\n");
if(pAPriv){
	pAPriv->pScrn=pScrn;
	pAPriv->Port[0].pAdaptor=pAPriv;
	pAPriv->Port[0].overlay_pad=-1;
	pAPriv->Port[0].overlay_scale_cntl=0x04000001L;
	pAPriv->Port[0].video_format=0xB000BL;
	pAPriv->Port[0].overlay_key=0x010203L;
	pAPriv->Port[0].overlay_id=0;
	pAPriv->Port[0].overlay_width=0;
	pAPriv->Port[0].overlay_height=0;
	pAPriv->Port[0].current_buffer=1;
	pAPriv->Port[0].capture_config=0;
	  /* below are hardcoded and present room for expansion.. */
	outm(OVERLAY_EXCLUSIVE_HORZ,0); /* disable exclusive mode */
	outm(OVERLAY_VIDEO_KEY_MSK,0xffff);
	outm(OVERLAY_KEY_CNTL,  (5<<4));
	outm(OVERLAY_GRAPHICS_KEY_CLR,pAPriv->Port[0].overlay_key);
	switch(pScrn->depth){
		case 8:
			outm(OVERLAY_GRAPHICS_KEY_MSK,0xff);
			break;
		case 15:
			outm(OVERLAY_GRAPHICS_KEY_MSK,0x7fff);
			break;
		case 16:
			outm(OVERLAY_GRAPHICS_KEY_MSK,0xffff);
			break;
		case 24:
			outm(OVERLAY_GRAPHICS_KEY_MSK,0xffffff);
			break;
		case 32:
			outm(OVERLAY_GRAPHICS_KEY_MSK,0xffffffff);
			break;
		
		}
	outm(SCALER_H_COEFF0,0x00002000);

	outm(SCALER_COLOUR_CNTL,  0x00101000);
  	outm(SCALER_H_COEFF0   , 0x00002000);
  	outm(SCALER_H_COEFF1   , 0x0D06200D);
  	outm(SCALER_H_COEFF2   , 0x0D0A1C0D);
  	outm(SCALER_H_COEFF3   , 0x0C0E1A0C);
  	outm(SCALER_H_COEFF4   , 0x0C14140C);
	
	outm(VIDEO_FORMAT, pAPriv->Port[0].video_format);
	
	outm(CAPTURE_CONFIG, pAPriv->Port[0].capture_config);
	m64_pollscreen(pScrn,pAPriv,0);
	}
switch(pATI->Chip){
	case ATI_CHIP_MOBILITY:
		pAPriv->hw_type=HW_RAGEPRO;
		break;
	default:
		pAPriv->hw_type=HW_UNSUPPORTED;
	}
return pAPriv;
}

void AtiVideoUninit(ScrnInfoPtr pScrn)
{
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiVideoUninit\n");
}

void AtiVideoReset(ScrnInfoPtr pScrn)
{
xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"AtiVideoReset\n");
}

void AtiVideoInit(ScreenPtr pScreen)
{
ScrnInfoPtr pScrn=xf86Screens[pScreen->myNum];
ATIPtr	pATI=ATIPTR(pScrn);
XF86VideoAdaptorRec VAR[2];
XF86VideoAdaptorPtr VARPtrs[2];
DevUnion Private[2];
AdaptorPrivPtr pAPriv;
int i;

xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"Initializing Xv driver\n");
pATI->XvEnabled=FALSE;

if (!(pAPriv = AdaptorPrivInit(pScrn))){
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Xv driver initialization failed\n");
	return;
    	}

if(pAPriv->hw_type==HW_UNSUPPORTED){
	xf86DrvMsgVerb(pScrn->scrnIndex,X_ERROR,1,"Xv: chipset is not supported\n");
	return;
	}



memset(VAR, 0, sizeof(VAR));

for (i = 0; i < 2; i++) {
	Private[i].ptr = (pointer) &pAPriv->Port[i];
	VARPtrs[i] = &VAR[i];

	VAR[i].type = (i == 0) ? XvInputMask : XvOutputMask;
	VAR[i].type |= XvWindowMask | XvVideoMask | XvStillMask | XvImageMask;
	VAR[i].name = "Ati";
	VAR[i].flags=VIDEO_OVERLAID_IMAGES;

	if (i == 0) {		
		VAR[0].pAttributes=ati_attributes_port0;
		VAR[0].nAttributes=NUM_ATI_ATTRIBUTES_PORT0;
	    	VAR[0].nEncodings = sizeof(InputVideoEncodings) / sizeof(InputVideoEncodings[0]);
	    	VAR[0].pEncodings = InputVideoEncodings;
	    	VAR[0].nFormats = sizeof(InputVideoFormats) / sizeof(InputVideoFormats[0]);
	    	VAR[0].pFormats = InputVideoFormats;
		VAR[0].pImages=ImagesG;
		VAR[0].nImages=NUM_IMAGES_G;
		} else {
	    	Bool expose = FALSE;

/*
	    xf86GetOptValBool(OutputOptions, OPTION_EXPOSE, &expose);
	    if (expose)	VAR[1].flags = VIDEO_EXPOSE;
*/

	    	VAR[1].nEncodings = sizeof(OutputVideoEncodings) / sizeof(OutputVideoEncodings[0]);
	    	VAR[1].pEncodings = OutputVideoEncodings;
	    	VAR[1].nFormats = sizeof(OutputVideoFormats) / sizeof(OutputVideoFormats[0]);
	    	VAR[1].pFormats = OutputVideoFormats;
		}

	VAR[i].nPorts = 1;
	VAR[i].pPortPrivates = &Private[i];
	VAR[i].PutVideo = AtiPutVideo;
	VAR[i].PutStill = AtiPutStill;
	VAR[i].GetVideo = AtiGetVideo;
	VAR[i].GetStill = AtiGetStill;
	VAR[i].StopVideo = AtiStopVideo;
	VAR[i].PutImage=AtiPutImage;
	VAR[i].ReputImage=AtiReputImage;
	VAR[i].QueryImageAttributes=AtiQueryImageAttributesG;
	VAR[i].SetPortAttribute = AtiSetPortAttribute;
	VAR[i].GetPortAttribute = AtiGetPortAttribute;
	VAR[i].QueryBestSize = AtiQueryBestSize;
    }

if (!xf86XVScreenInit(pScreen, VARPtrs, 1)){  /* register only the first, 
						 input port - the output is
						 not ready yet */
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Xv initialization failed\n");
	return;
	}


xvEncoding   = MAKE_ATOM(XV_ENCODING);
xvHue        = MAKE_ATOM(XV_HUE);
xvSaturation = MAKE_ATOM(XV_SATURATION);
xvBrightness = MAKE_ATOM(XV_BRIGHTNESS);
xvContrast   = MAKE_ATOM(XV_CONTRAST);

xv_format=MAKE_ATOM(XV_FORMAT);
xv_color_key=MAKE_ATOM(XV_COLOR_KEY);

pATI->XvEnabled=TRUE;

xf86DrvMsgVerb(pScrn->scrnIndex,X_INFO,1,"Xv driver enabled\n");
}
