#include "Xarch.h"
#include "xaa.h"
#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"
#include "xf86Pci.h"
#include "xf86PciInfo.h"
#include "xf86Xinput.h"
#include "xf86fbman.h"
#include "xf86i2c.h"
#include "xf86xv.h"
#include "Xv.h"
#include "xf86Cursor.h"
#include "regionstr.h"
#include "xaalocal.h"

#include "r128.h"
#include "r128_reg.h"
#include "r128_video.h"
#include "r128_i2c.h"


void R128_I2CInit(PortPrivPtr pPPriv)
{
double nm;
R128InfoPtr   info = R128PTR(pPPriv->pAdaptor->pScrn);
R128PLLPtr    pll  = &info->pll;
ScrnInfoPtr	pScrn=pPPriv->pAdaptor->pScrn;
R128InfoPtr	pATI=R128PTR(pScrn);
R128MMIO_VARS();

nm=(pll->reference_freq * 10000.0 )/(4.0 * R128_CLOCK_FREQ);

for(pPPriv->r128_N=1; pPPriv->r128_N<255; pPPriv->r128_N++)
	if((pPPriv->r128_N * (pPPriv->r128_N-1)) > nm)break;
pPPriv->r128_M=pPPriv->r128_N-1;
pPPriv->r128_time=2*pPPriv->r128_N;

outm8(R128_I2C_CNTL_1+2, ((I2C_SEL | I2C_EN)>>16));
outm8(R128_I2C_CNTL_0+0,  (I2C_DONE | I2C_NACK | I2C_HALT | I2C_SOFT_RST | I2C_DRIVE_EN | I2C_DRIVE_SEL));

}


/******************************************************************************
 * R128_WaitForFifo - wait n empty FIFO entries.                              *
 *  Function: The FIFO contains upto 64 empty entries. The 'entries'          *
 *            value must be 1 to 64. This function implements the same        *
 *            timeout mechanism as the R128_WaitForIdle() function.           *
 *    Inputs: entries - number of entries spaces to wait for. Max - 64        *
 *   Outputs: NONE                                                            *
 ******************************************************************************/
void R128_WaitForFifo (PortPrivPtr pPPriv, CARD32 entries)
{
int count;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ScrnInfoPtr	pScrn=pAPriv->pScrn;
R128InfoPtr	pATI=R128PTR(pScrn);
R128MMIO_VARS();

  count=0;

    while ((inm(R128_GUI_STAT) & 0x00000FFF) < entries)
    {
    count++;
    usleep(1);
    if(count>1000)return;
    } 

    return;

} /* R128_WaitForFifo () */


/****************************************************************************
 *  I2C_WaitForAck (void)                                                   *
 *                                                                          *
 *  Function: polls the I2C status bits, waiting for an acknowledge or      *
 *            an error condition.                                           *
 *    Inputs: NONE                                                          *
 *   Outputs: I2C_DONE - the I2C transfer was completed                     *
 *            I2C_NACK - an NACK was received from the slave                *
 *            I2C_HALT - a timeout condition has occured                    *
 ****************************************************************************/
CARD8 I2C_WaitForAck (PortPrivPtr pPPriv)
{
CARD8 retval = 0;
CARD8 test;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ScrnInfoPtr	pScrn=pAPriv->pScrn;
R128InfoPtr	pATI=R128PTR(pScrn);
R128MMIO_VARS();

    usleep(1);
    do
    {
        retval = inm8(R128_I2C_CNTL_0);
        if (retval & I2C_HALT)
        {
            return (I2C_HALT);
        }
        if (retval & I2C_NACK)
        {
            return (I2C_NACK);
        }
        test = retval && (CARD8)I2C_DONE;
    }
    /*while ((retval & I2C_DONE) != 0);*/
    while (test == 0);

    return (I2C_DONE);
}


/****************************************************************************
 *  I2C_Set (I2C_PACKET *Packet)                                           *
 *  Function: Sets data at an I2C address                                   *
 *    Inputs: I2C_PACKET *Packet - pointer to I2C Packet structre          *
 *   Outputs: I2C_DONE - success, the data was accepted by the slave        *
 *            I2C_NACK - the data was not accepted                          *
 *            I2C_HALT - a timeout condition has occured                    *
 ****************************************************************************/
CARD8 I2C_Set (PortPrivPtr pPPriv, I2C_PACKET *Packet)
{
CARD8 loop;
CARD32 i2c_cntl_0, i2c_cntl_1;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ScrnInfoPtr	pScrn=pAPriv->pScrn;
R128InfoPtr	pATI=R128PTR(pScrn);
R128MMIO_VARS();

    R128_WaitForFifo (pPPriv, 4 + Packet->number_of_bytes);

    /* Clear the status bits of the I2C Controller */
    outm(R128_I2C_CNTL_0, I2C_DONE | I2C_NACK | I2C_HALT | I2C_SOFT_RST);

    /* Write the address into the buffer first */
    outm(R128_I2C_DATA, (CARD32) Packet->address & ~(1));

    /* Write Value into the buffer */
    for (loop = 0; loop < Packet->number_of_bytes; loop++)
    {
        outm8(R128_I2C_DATA, Packet->data[loop]);
    }

    i2c_cntl_1 = (pPPriv->r128_time << 24) | I2C_EN | I2C_SEL |
                  Packet->number_of_bytes | 0x00000100;
    outm(R128_I2C_CNTL_1, i2c_cntl_1);

    /* set I2C_CNTL_0 */
    i2c_cntl_0 = (pPPriv->r128_N << 24) | (pPPriv->r128_M << 16) |
                  I2C_GO | (Packet->start << 8) | (Packet->stop << 9) | I2C_DRIVE_EN;
    outm(R128_I2C_CNTL_0, i2c_cntl_0);

    /* wait for finishing of I2C operation */
    while (inm8(R128_I2C_CNTL_0+1) & (I2C_GO>>8));

    loop = I2C_WaitForAck (pPPriv);

    return (loop);
} /* I2C_Set ()... */

/****************************************************************************
 *  I2C_Get (I2C_PACKET *Packet)                                           *
 *                                                                          *
 *  Function: Gets data from an I2C address                                 *
 *    Inputs: I2C_PACKET *Packet - pointer to I2C Packet structre          *
 *   Outputs: I2C_DONE - success, the data was received from the slave      *
 *            I2C_NACK - the data was not received                          *
 *            I2C_HALT - a timeout condition has occured                    *
 ****************************************************************************/
CARD8 I2C_Get (PortPrivPtr pPPriv, I2C_PACKET *Packet)
{
CARD8 loop, retval;
CARD32 i2c_cntl_0, i2c_cntl_1;
AdaptorPrivPtr pAPriv = pPPriv->pAdaptor;
ScrnInfoPtr	pScrn=pAPriv->pScrn;
R128InfoPtr	pATI=R128PTR(pScrn);
R128MMIO_VARS();

    R128_WaitForFifo (pPPriv, 4 + Packet->number_of_bytes);

    outm(R128_I2C_CNTL_0, 0x00000027);

    /* set address byte */
    outm(R128_I2C_DATA, Packet->address | 1);

    /* receive num_data byte(s) and transmit 1 address byte */
    i2c_cntl_1 = (pPPriv->r128_time << 24) | I2C_EN | I2C_SEL |
                  Packet->number_of_bytes | 0x00000100;
    outm(R128_I2C_CNTL_1, i2c_cntl_1);

    /* set I2C_CNTL_0 */
    i2c_cntl_0 = (pPPriv->r128_N << 24) | (pPPriv->r128_M << 16) |
                  I2C_GO | (Packet->start << 8) | (Packet->stop << 9) | I2C_DRIVE_EN | I2C_RECEIVE;
    outm(R128_I2C_CNTL_0, i2c_cntl_0);

    /* wait for finishing of I2C operation */
    while (inm8(R128_I2C_CNTL_0+1) & (I2C_GO>>8));

    retval = I2C_WaitForAck (pPPriv);

    /* Wait while the buffer registers are filled. */
     usleep(1);

    for (loop = 0; loop < Packet->number_of_bytes; loop++)
    {
        /* read the data register */
        R128_WaitForFifo (pPPriv,1);
        if ((retval == I2C_HALT) || (retval == I2C_NACK))
        {
            Packet->data[loop] = 0xFF;
        }
        else
        {
            Packet->data[loop] = inm8(R128_I2C_DATA) & 0xff;
        }
    }

    return (I2C_DONE);
} /* I2C_Get */


Bool I2CProbeAddress(PortPrivPtr pPPriv, CARD8 addr)
{
I2C_PACKET packet;
CARD8 data[3];

packet.number_of_bytes=1;
packet.address=addr;
packet.data=data;
packet.start=1;
packet.stop=1;


if(!(I2C_Get(pPPriv, &packet)==I2C_DONE))return FALSE;
if(data[0]==0xff)return FALSE;
return TRUE;
}
