#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/videodev.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/Xvlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/keysym.h>
#define _GNU_SOURCE
#include <getopt.h>

/* ok the following constant should be by right included thru in Xvlib.h */
#ifndef GUID_YUV12_PLANAR
#define GUID_YUV12_PLANAR 0x32315659
#endif

#ifndef GUID_YUY2_PACKED
#define GUID_YUY2_PACKED 0x32595559
#endif

#ifndef GUID_UYVY_PACKED
#define GUID_UYVY_PACKED 0x59565955
#endif

int video_fd=-1;
char *device_name="/dev/video";
struct video_capability vcap;
struct video_picture vpic;
struct video_window vwin;

Display *display=NULL;
Window window=(Window)NULL;
long width,height;
long device_width,device_height;
int connection=-1;
XvImage *xv_image=NULL;
XvAdaptorInfo *info;
XShmSegmentInfo shminfo1;
long format=0;
GC gc;
Atom xv_color_key;
long background=0x010203;

unsigned char *frame_data[2]={NULL,NULL};
long frame_length;
long frame_free;
int frame=0;
int adaptor=-1;

int freeze=0;
int average=0;
int fps;
int res;


static struct option long_options[]={
	{"device",1,NULL,0},
	{"fps",1,NULL,0},
	{"res",1,NULL,0},
	{"help",0,NULL,0},
	{NULL,0,0,0}
	};

void get_options(int argc,char *argv[])
{
int option_index=0;
fps=10;
res=0;

while(getopt_long(argc,argv,"",long_options,&option_index)>=0){
	if(optarg){
		switch(option_index){ 
			/* case values must match long_options */
			case 0:
				device_name=strdup(optarg);
				break;
			case 1:
				fps=atoi(optarg);
				break;
			case 2: 
				res=atoi(optarg);
				break;
			}
		}
	if(option_index==3){
		printf( "\n"
		        "        %s - Xv video monitor\n\n"
			"Usage:\n"
			"        %s [--fps=10] [--res=[0,1]] [--device=/dev/video] [--help]\n"
			"             --fps    - frames per second. higher rates might be incompatible\n"
			"                        with higher resolutions\n"
			"             --res    - resolution 0 means 352x288, 1 means 640x480\n"
			"             --device - specifies video4linux device to use\n"
			"             --help   - prints this message\n",argv[0],argv[0]);
		exit(0);
		}
	}

}


void set_frame_length(long size)
{
if(frame_data[0]!=NULL)free(frame_data);
if(frame_data[1]!=NULL)free(frame_data);
frame_length=size;
fprintf(stderr,"Setting frame size to %ld kb\n",size/1024);
frame_data[0]=malloc(frame_length+20);
frame_data[1]=malloc(frame_length+20);
frame_free=0;
}

void display_frame(void)
{
if(adaptor>=0){
	xv_image=XvCreateImage(display,info[adaptor].base_id,format,frame_data[!frame],device_width,device_height);
	XvPutImage(display,info[adaptor].base_id,window,gc,xv_image,
		0,0,device_width,device_height,
		0,0,width,height);
	xv_image=NULL;
	} 
}

void average_frame(void)
{
long i;
unsigned char *from;
unsigned char *to;
short int diff1,diff2;
from=frame_data[frame];
to=frame_data[!frame];
diff1=0;
for(i=0;i<frame_length;i++){
	diff2=to[i]-from[i];
	if((diff1>0)+(diff2>0)==1)to[i]=from[i]+((to[i]-from[i])>>3);
		else to[i]=from[i]+((to[i]-from[i])>>3);
	diff1=diff2;
	}
}

void diff_frame(void)
{
long i;
unsigned char *from;
unsigned char *to;
from=frame_data[frame];
to=frame_data[!frame];
for(i=0;i<frame_length;i++){
	to[i]=(from[i]+to[i])>>1;
	}

}

void QueryXv(void)
{
int num_adaptors;
int num_formats;
XvImageFormatValues *formats;
int i,j;
char xv_name[5];

XvQueryAdaptors(display,DefaultRootWindow(display),&num_adaptors,&info);

for(i=0;i<num_adaptors;i++){
	formats=XvListImageFormats(display,info[i].base_id,&num_formats);
	for(j=0;j<num_formats;j++){
		xv_name[4]=0;
		memcpy(xv_name,&formats[j].id,4);
		fprintf(stderr,"   0x%x %s %s\n",formats[j].id,xv_name,(formats[j].format==XvPacked)?"packed":"planar");
		if(formats[j].id==format){
			if(adaptor<0)adaptor=i;
			}
		}
	}
if(adaptor<0)
	fprintf(stderr,"No suitable Xv adaptor found");		
}

int SetFrameRate(int fps)
{
long flags;
flags=vwin.flags;
vwin.flags=(vwin.flags & ~0x3f) | (fps & 0x3f);
if(ioctl(video_fd,VIDIOCSWIN,&vwin)<0){
	vwin.flags=flags;
	return -1;
	} else {
	return 0;
	}
}

void InitShmXv(void)
{
if(adaptor<0)return;

xv_image=(XvImage *) XvShmCreateImage(display,info[adaptor].base_id,
	GUID_YUY2_PACKED,0,device_width,device_height,(char *)&shminfo1);

shminfo1.shmid=shmget(IPC_PRIVATE,xv_image->data_size,IPC_CREAT|0777);
shminfo1.shmaddr=shmat(shminfo1.shmid,0,0);
xv_image->data=shminfo1.shmaddr;
shminfo1.readOnly=0;
if(!XShmAttach(display,&shminfo1)){
	fprintf(stderr,"Cannot attach shared memory\n");
	xv_image=NULL;
	return;
	}
}

int main(int argc,char *argv[])
{
XEvent xev;
fd_set r_fds;
struct timeval tv;
long count;
XGCValues xgcv;

device_width=352;
device_height=288;

get_options(argc,argv);

video_fd=open(device_name,O_RDONLY);
if(video_fd<0){
	perror("open");
	exit(-1);
	}
if(ioctl(video_fd,VIDIOCGCAP,&vcap)<0){
	perror("getting video capabilities");
	exit(-1);
	}

if(ioctl(video_fd,VIDIOCGPICT,&vpic)<0){
	perror("error getting picture formats");
	exit(-1);
	}
	
fprintf(stderr,"Detected \"%s\" [%s%s ] with %d channels and %d audio devices\n",vcap.name,
		vcap.type & VID_TYPE_CAPTURE ? " capture":"",
		vcap.type & VID_TYPE_TUNER ? " tuner":"",
		vcap.channels,vcap.audios);
fprintf(stderr,"\tmaximum capture window is %dx%d\n"
		"\tminimum capture window is %dx%d\n",
		vcap.maxwidth,vcap.maxheight,
		vcap.minwidth,vcap.minheight);

fprintf(stderr,"Current parameters:\n"
	 	"\tbrightness: %d\n"
		"\thue: %d\n"
		"\tcolour: %d\n"
		"\tcontrast: %d\n"
		"\twhiteness: %d\n"
		"\tdepth: %d\n"
		"\tpalette: 0x%x\n",
		vpic.brightness,vpic.hue,vpic.colour,vpic.contrast,vpic.whiteness,vpic.depth,vpic.palette);

if(!(vcap.type & VID_TYPE_CAPTURE)){
	fprintf(stderr,"No capture support present - exiting\n");
	exit(-1);
	}

	
if(ioctl(video_fd,VIDIOCGWIN,&vwin)<0){
	perror("error getting window parameters");
	exit(-1);
	}
vwin.x=0;
vwin.y=0;
vwin.width=device_width;
vwin.height=device_height;
if(ioctl(video_fd,VIDIOCSWIN,&vwin)<0){
	perror("error setting window parameters");
	exit(-1);
	}

SetFrameRate(15);

display=XOpenDisplay(getenv("DISPLAY"));
if(display==NULL){
	fprintf(stderr,"Could not open display \"%s\"\n",getenv("DISPLAY"));
	exit(-1);
	}


vpic.hue=0xffff/2;
vpic.brightness=0xffff/2;
vpic.colour=0xffff/2;
vpic.contrast=0xffff/2;
vpic.whiteness=0xffff/2;

vpic.palette=VIDEO_PALETTE_YUV420P;
vpic.palette=VIDEO_PALETTE_UYVY; /* this one is broken for my cam */
vpic.palette=VIDEO_PALETTE_YUYV;
if(ioctl(video_fd,VIDIOCSPICT,&vpic)<0){
	perror("error setting picture parameters, perhaps YUYV is not supported");	
	exit(-1);
	}
format=GUID_YUV12_PLANAR;
format=GUID_YUY2_PACKED;
/* format=GUID_UYVY_PACKED; */
QueryXv();
fprintf(stderr,"Using 0x%08x\n", format);

if(adaptor<0){
	exit(-1);
	}

width=device_width;
height=device_height;
window=XCreateSimpleWindow(display,DefaultRootWindow(display),1,1,width,height,0,
	WhitePixel(display,DefaultScreen(display)),
	background);
	
XSelectInput(display,window,StructureNotifyMask|KeyPressMask);
XMapWindow(display,window);
connection=ConnectionNumber(display);

gc=XCreateGC(display,window,0,&xgcv);
xv_color_key=XInternAtom(display,"XV_COLORKEY",False);
XvSetPortAttribute(display,info[adaptor].base_id,xv_color_key,background);

switch(format){
	case GUID_YUV12_PLANAR:
		set_frame_length(device_width*device_height*3/2);
		break;
	case GUID_YUY2_PACKED:
	case GUID_UYVY_PACKED:
		set_frame_length(device_width*device_height*2);
		break;
	default:
		fprintf(stderr,"Unknown format set (internal error)\n");
		exit(255);
	}

while(1){
	FD_ZERO(&r_fds);
	FD_SET(connection,&r_fds);
	FD_SET(video_fd,&r_fds);
	tv.tv_usec=1;
	tv.tv_sec=1;
	select(connection>video_fd?connection+1:video_fd+1,
		&r_fds,NULL,NULL,&tv);
		
	while(XPending(display)>0){
		XNextEvent(display,&xev);
		switch(xev.type){
			case ConfigureNotify:
				width=xev.xconfigure.width;
				height=xev.xconfigure.height;
				display_frame();
				break;
			case KeyPress:
				switch(XKeycodeToKeysym(display,xev.xkey.keycode,0)){
					case XK_space: freeze=!freeze;
						break;
					case XK_a:
					case XK_A:
						average=!average;
						break;
					case XK_q:
					case XK_Q:
						exit(0);
						break;
					case XK_comma:
					case XK_less:
						width=width/2;
						height=height/2;
						XResizeWindow(display,window,width,height);
						display_frame();
						break;
					case XK_period:
					case XK_greater:
						width=2*width;
						height=2*height;
						XResizeWindow(display,window,width,height);
						display_frame();
						break;
					}
				break;
			default:
				printf(".");
			}
		}
	if(FD_ISSET(video_fd,&r_fds)){
		count=read(video_fd,frame_data[frame]+frame_free,frame_length-frame_free);
		if(count>0){
			frame_free+=count;			
			if(frame_free==frame_length){
				if(!freeze)frame=!frame;
				if(average)average_frame();
				display_frame();
				XFlush(display);
				frame_free=0;
				}
			} else 
		if(count==0){
			fprintf(stderr,"device stopped capturing..\n");
			XUnmapWindow(display,window);
			XFlush(display);
			exit(0);
			}
		}	
	}


exit(0);
}
