/**************************************************************************\
 gatos (General ATI TV and Overlay Software)

  Project Coordinated By Insomnia (Steaphan Greene)
  (insomnia@core.binghamton.edu)

  Copyright (C) 1999 Steaphan Greene, yvind Aabling, Octavian Purdila, 
	Vladimir Dergachev, Christian Lupien and Daniel Caujolle-Bert.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

\**************************************************************************/

#define GATOS_A_SOUND_C 1

#include "gatos.h"
#include "a_sound.h"
#include "board.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <math.h>

static int setmixer=0 ;
static int mixers_available = 0;
static int card = -1, device = -1;
char mixer_device[256];

//#define ASND_DEBUG 1

/* ------------------------------------------------------------------------ */

#define aserr(FMT,ARGS...)    fprintf(stderr,FMT,##ARGS)

/* ------------------------------------------------------------------------ */
/*  			       Private Routines                             */
/* ------------------------------------------------------------------------ */

void disable_sound(void) {

  gatos.audiotype &= DISABLE_MIXER;
  gatos.mixerdev = -1 ;
  mixers_available = 0;
}

/* ------------------------------------------------------------------------ */

int hw_info(int info_card) {
  int err;
  snd_ctl_t *hw_handle;
  snd_ctl_hw_info_t hw_info;
  
  if((err = snd_ctl_open(&hw_handle, info_card)) < 0) {
    aserr("GATOS: hw_info() failed: %s\n", snd_strerror(err));
    snd_ctl_close(hw_handle);
    RETURN0;
  }
  
  if((err = snd_ctl_hw_info(hw_handle, &hw_info)) < 0) {
    aserr("GATOS: snd_ctl_hw_info() failed: %s\n", snd_strerror(err));
    snd_ctl_close(hw_handle);
    RETURN0;
  }
  
  aserr("\n");
  aserr(" snd_ctl_hw_info struct:\n");
  aserr(" -----------------------\n");
  aserr("  Type of card:              SND_CARD_TYPE_(0x%x)\n",
	hw_info.type);
  aserr("  Count of hw dependent dev: %i\n", hw_info.hwdepdevs);
  aserr("  Count of PCM devices:      %i\n", hw_info.pcmdevs);
  aserr("  Count of MIXERS devices:   %i\n", hw_info.mixerdevs);
  aserr("  Count of raw MIDI devices: %i\n", hw_info.mididevs);
  aserr("  Count of timer devices:    %i\n", hw_info.timerdevs);
  aserr("  ID of card:                %s\n", hw_info.id);
  aserr("  Abbreviation for sndcard:  %s\n", hw_info.abbreviation);
  aserr("  Short name of sndcard:     %s\n", hw_info.name);
  aserr("  Name+info about sndcard:   %s\n", hw_info.longname);
  aserr("  Reserved:                  %s\n", hw_info.reserved);
  aserr(" ---\n");
  
  snd_ctl_close(hw_handle);
  
  return 1;
}

/* ------------------------------------------------------------------------ */

int volume_to_percent(int volume, int vol_min, int vol_max) {
  int range = vol_max - vol_min;
  int return_percent;
  
  if(range <= 0) return 0;
  volume -= vol_min;
  return_percent = rint((double)volume / (double)range * 100);

  return return_percent;
}

/* ------------------------------------------------------------------------ */

int percent_to_volume(int percent, int vol_min, int vol_max) {
  int range = vol_max - vol_min;
  int return_volume;
  
  if(range <= 0) return 0;
  percent -= vol_min;
  return_volume = rint((double)percent * (double)range / 100);

  return return_volume;
}
/* ------------------------------------------------------------------------ */

int set_volume(char *mixer_grp, int volume) {
  int err;
  snd_mixer_t *handle;
  snd_mixer_gid_t gid;
  snd_mixer_group_t group;

  if(SOUNDINFO)
    aserr("set_volume(): mixer_grp=%s, volume=%i\n", mixer_grp, volume);

  bzero(&gid, sizeof(gid));
  bzero(&group, sizeof(group));

  strcpy(gid.name, mixer_grp);
  
  if((err = snd_mixer_open(&handle, card, device)) < 0) {
    aserr(_("GATOS: mixer %i/%i open error: %s\n"), card, device, 
	  snd_strerror(err));
    return (err);
  }

  group.gid = gid;
  
  if(snd_mixer_group_read(handle, &group)<0) {
    aserr(_("GATOS: unable to read group '%s',%i: %s\n"), gid.name, gid.index, 
	  snd_strerror(err));
    snd_mixer_close(handle);
    return -1;
  }

  group.volume.values[0] = group.volume.values[1] = 
    percent_to_volume(volume, group.min, group.max);

  if(gatos.mute) {
    if(SOUNDINFO)
      aserr("gatos is muted\n");
    group.mute = group.channels;
    /*      //    group.mute &= (1<<1); */
  }
  else {
    if(SOUNDINFO)
      aserr("gatos is unmuted\n");
    group.mute = 0;
    /*      //    group.mute &= (0<<1); */
  }
  
  if(snd_mixer_group_write(handle, &group) < 0) {
    aserr(_("GATOS: unable to write group '%s',%i: %s\n"), gid.name, gid.index, 
	  snd_strerror(err));
    snd_mixer_close(handle);
    return -1;
  }
  
  snd_mixer_close(handle);
  
  RETURN0;
}
/* ------------------------------------------------------------------------ */

int get_volume(char *mixer_grp) {
  int err, chan;
  snd_mixer_t *handle; 
  snd_mixer_gid_t gid;
  snd_mixer_group_t group;
  int current_volume = 0;

  if(SOUNDINFO)
    aserr("get_volume(): mixer_grp=%s\n", mixer_grp);

  bzero(&gid, sizeof(gid));
  strcpy(gid.name, mixer_labels[gatos.mixerdev]);

  bzero(&group, sizeof(group));
  group.gid = gid;

  if((err = snd_mixer_open(&handle, card, device)) < 0) {
    aserr(_("GATOS: mixer %i/%i open error: %s\n"), card, device, 
	  snd_strerror(err));
    return -1;
  }
  
  if((err = snd_mixer_group_read(handle, &group)) < 0) {
    aserr(_("GATOS: mixer %i/%i group error: %s\n"), card, device, 
	  snd_strerror(err));
    return -1;
  }

  if(group.channels) {
    if(group.channels == SND_MIXER_CHN_MASK_MONO) {
      if(SOUNDINFO) {
	aserr("Mono: [%s] [%%%i]\n",
	      group.mute & SND_MIXER_CHN_MASK_MONO ? "mute" : "on",
	      volume_to_percent(group.volume.names.front_left, 
				group.min, group.max));
      }
      current_volume = volume_to_percent(group.volume.names.front_left, 
					 group.min, group.max);
    } 
    else {
      for(chan = 0; chan <= SND_MIXER_CHN_LAST; chan++) {
	if(!(group.channels & (1<<chan)))
	  continue;
	if(SOUNDINFO) {
	  aserr("Stereo: chan=%i %i [%i] [%s]=%i\n", chan, 
		group.volume.values[chan], 
		volume_to_percent(group.volume.values[chan], 
				  group.min, group.max),
		(group.mute & (1<<chan) ? "mute":"unmuted"), group.mute);
	}
	current_volume = volume_to_percent(group.volume.values[chan], 
					   group.min, group.max);
      }
    }
  }

  return current_volume;
}

/* ------------------------------------------------------------------------ */
/*  			      Public Routines                               */
/* ------------------------------------------------------------------------ */

int sound_init(void) {
  int i, err, idx;
  snd_mixer_t *handle;
  snd_mixer_info_t mixer_info;
  snd_mixer_groups_t groups;
  snd_mixer_gid_t *group;

#ifdef ASND_DEBUG
  aserr("sound_init()\n");
#endif 

  if(VERBOSE) aserr(_("ALSA sound support, version %d.%d.%d\n"),
		    SND_LIB_MAJOR, SND_LIB_MINOR, SND_LIB_SUBMINOR);

  card = snd_defaults_mixer_card();
  device = snd_defaults_mixer_device();

  if(card < 0 || device < 0) {
    aserr(_("GATOS: sound_init() failed, no soundcard or mixer available\n"));
    disable_sound();
  }

  if(SOUNDINFO) {
    aserr("Card(s) available:  %i\n", snd_cards());
    aserr("card=%d, device=%d\n", card, device);
    aserr("Bitmap of sndcards: 0x%08x\n", snd_cards_mask());
  }
  
  if(gatos.audiotype & SNDCARD_MIXER) { 
    
    if(SOUNDINFO) {
      (void *) hw_info(card);
    }

    if((err = snd_mixer_open(&handle, card, device)) < 0) { 
      aserr(_("GATOS: snd_mixer_open() failed: %s\n"), snd_strerror(err));
      disable_sound();
    } 
    
    bzero(&groups, sizeof(groups));

    if((err = snd_mixer_groups(handle, &groups)) < 0) {
      aserr(_("GATOS: mixer %i/%i groups error: %s\n"), card, device, 
	    snd_strerror(err));
      disable_sound();
    }
    
    groups.pgroups = (snd_mixer_gid_t *) malloc(groups.groups_over * 
						sizeof(snd_mixer_eid_t));
    
    if(!groups.pgroups) {
      aserr(_("GATOS: sound_init(): not enough memory\n"));
      disable_sound();
    }
    
    groups.groups_size = groups.groups_over;
    groups.groups_over = groups.groups = 0;
    
    if((err = snd_mixer_groups(handle, &groups)) < 0) {
      aserr(_("GATOS: mixer %i/%i groups (2) error: %s\n"), card, device, 
	    snd_strerror(err));
      disable_sound();
    }
    
    for(idx = 0; idx < groups.groups; idx++) {
      group = &groups.pgroups[idx];

      if(SOUNDINFO)
	aserr(" - Group: %s, %i\n", group->name, group->index);

      mixer_labels[idx] = (char *) malloc(strlen(group->name)+1);
      strncpy(mixer_labels[idx], group->name, strlen(group->name));
    }
    free(groups.pgroups);
    
    if((err = snd_mixer_info(handle, &mixer_info)) < 0) {
      aserr(_("GATOS: snd_mixer_info(): %s\n"), snd_strerror(err));
      snd_mixer_close(handle);
      disable_sound();
    }
    
    snprintf(mixer_device, 256, "/dev/snd/mixerC%iD%i", card, device);

    if(SOUNDINFO) {
      aserr("\n");
      aserr(" Mixer device:               %s\n", mixer_device);
      aserr("\n");
      aserr(" snd_mixer_info struct:\n");
      aserr(" ----------------------\n");
      aserr("  Type of sndcard:           SND_CARD_TYPE_XXXX=0x%08x\n",
	    mixer_info.type);
      aserr("  Device device attribs:     SND_MIXER_ATTR_(0x%x)\n",
	    mixer_info.attrib);
      aserr("  ID of this mixer:          %s\n", mixer_info.id);
      aserr("  Device name:               %s\n", mixer_info.name);
      aserr("  Count of elements:         %d\n", mixer_info.elements);
      aserr("  Count of element groups:   %d\n", mixer_info.groups);
      aserr("  Reserved:                  %s\n", mixer_info.reserved);
      aserr(" ---\n");
    }

    mixers_available = mixer_info.groups;

    gatos.mixerfd = snd_mixer_file_descriptor(handle);
        
    if(gatos.mixerfd < 0)
      aserr(_("GATOS: open(%s) failed\n"), mixer_device);
    
    gatos.mixerdev = -1;

    /* Default to line-in */
    if(gatos.mixerfd >= 0)
      for(i = 1; i < mixers_available; i++)
	if(!strncasecmp("line", mixer_labels[i], 4)) {
	  gatos.mixerdev = i;
	  if(SOUNDINFO)
	    aserr("mixer forced to %i, %s, gatos.mixerdev=%i\n", i, 
		  mixer_labels[i], gatos.mixerdev);
	}
    
    if(gatos.mixerdev <= 0 || 
       ((gatos.oldvol = get_volume(mixer_labels[gatos.mixerdev])) < 0 ))
      if(gatos.mixerfd >= 0) {
	aserr(_("GATOS: mixer read failed.\n"));
	gatos.audiotype &= DISABLE_MIXER ; 
      }

    snd_mixer_close(handle);
    
#ifdef ASND_DEBUG
    aserr("   --- gatos.volume=%i\n",   gatos.volume);
    aserr("   --- gatos.oldvol=%i\n",   gatos.oldvol);
    aserr("   --- gatos.mute=%i\n",     gatos.mute);
    aserr("   --- gatos.stereo=%i\n",   gatos.stereo);
    aserr("   --- gatos.sap=%i\n",      gatos.sap);
    aserr("   --- gatos.mixerdev=%i\n", gatos.mixerdev);
#endif
  } 
  
  if(gatos.audiotype==BRDCARD_MIXER) {
    gatos_setvolume(gatos.volume);
  }
  
  RETURN0 ; 
}

/* ------------------------------------------------------------------------ */

int sound_restore(void) {
  int i;

  if(gatos.audiotype&SNDCARD_MIXER) {
    if(set_volume(mixer_labels[gatos.mixerdev], gatos.oldvol) < 0) {
      aserr(_("GATOS: cannot restore old volume %i for mixer %s\n"), gatos.oldvol,
	    mixer_labels[gatos.mixerdev]);
    }
  }
  if(gatos.audiotype&BRDCARD_MIXER) {
    /*board_setvol(gatos.oldvol);*/
  }

#ifdef ASND_DEBUG
  aserr("   --- gatos.volume=%i\n",   gatos.volume);
  aserr("   --- gatos.oldvol=%i\n",   gatos.oldvol);
  aserr("   --- gatos.mute=%i\n",     gatos.mute);
  aserr("   --- gatos.stereo=%i\n",   gatos.stereo);
  aserr("   --- gatos.sap=%i\n",      gatos.sap);
  aserr("   --- gatos.mixerdev=%i\n", gatos.mixerdev);
  aserr(" Freeing Memory\n");
#endif 

  for(i = mixers_available; i > 0; i--) free(mixer_labels[i]);

  RETURN0 ; 
}

/* ------------------------------------------------------------------------ */

int sound_setvol(int value) {
  int err;

  if(SOUNDINFO)
    aserr("sound_setvol(): volume=%d\n", value);

  if(gatos.audiotype&SNDCARD_MIXER && setmixer) {
    if(gatos.mixerdev > 0)
      if((err = set_volume(mixer_labels[gatos.mixerdev], value)) < 0) {
	aserr(_("GATOS: unable to set volume %i for card %i\n"), value, card);
	gatos.audiotype &= DISABLE_MIXER ; 
      }
  }
  
#ifdef ASND_DEBUG
  aserr("sound_setvol(): value=%i, set=%i\n", 
	value, get_volume(mixer_labels[gatos.mixerdev]));
#endif

  if(gatos.audiotype&BRDCARD_MIXER) {
    /*board_setvol(value);*/
  }

#ifdef ASND_DEBUG
  aserr("   --- gatos.volume=%i\n",   gatos.volume);
  aserr("   --- gatos.oldvol=%i\n",   gatos.oldvol);
  aserr("   --- gatos.mute=%i\n",     gatos.mute);
  aserr("   --- gatos.stereo=%i\n",   gatos.stereo);
  aserr("   --- gatos.sap=%i\n",      gatos.sap);
  aserr("   --- gatos.mixerdev=%i\n", gatos.mixerdev);
#endif
  
  RETURN0 ;
}

/* ------------------------------------------------------------------------ */

int sound_setmixer(int value) {

  if(SOUNDINFO)
    aserr("sound_setmixer(): mixer=%d = %s, avail=%d\n", value, 
	  mixer_labels[value], mixers_available);

  if(mixers_available > 0 || value > 0) {
    if(value > mixers_available) {
      aserr(_("GATOS: selected mixer is out of range %i/%i\n"), value, 
	    mixers_available);
      disable_sound();
      RETURN0;
    }
    
    gatos.mixerdev = value;
    setmixer = 1;
    
    if(gatos.mixerdev >= 0 &&
       ((gatos.oldvol = get_volume(mixer_labels[gatos.mixerdev])) < 0 )) {
      aserr(_("GATOS: mixer set failed.\n"));
      gatos.audiotype &= DISABLE_MIXER;
    }
    else 
      gatos.audiotype |= SNDCARD_MIXER;
    
  }

#ifdef ASND_DEBUG
    aserr("   --- gatos.volume=%i\n",   gatos.volume);
    aserr("   --- gatos.oldvol=%i\n",   gatos.oldvol);
    aserr("   --- gatos.mute=%i\n",     gatos.mute);
    aserr("   --- gatos.stereo=%i\n",   gatos.stereo);
    aserr("   --- gatos.sap=%i\n",      gatos.sap);
    aserr("   --- gatos.mixerdev=%i\n", gatos.mixerdev);
#endif
  
  RETURN0;
}

/* ------------------------------------------------------------------------ */

int sound_nummixers(void) {

  if(SOUNDINFO)
    aserr("sound_nummixers(): mixers avail.=%d\n", mixers_available);

  return mixers_available;
}

/* ------------------------------------------------------------------------ */

char* sound_mixername(int value) {

  if(SOUNDINFO)
    aserr("sound_mixername(): want=%d, is=%s\n", value, 
	  mixer_labels[value]);

  return mixer_labels[value];
}

/* ------------------------------------------------------------------------ */

char** sound_mixernames(void) {

  return mixer_labels;
}

/* ------------------------------------------------------------------------ */

char** sound_mixerlabels(void) {

  return mixer_labels;
}

/* ------------------------------------------------------------------------ */
