/*
To compile use:
  cc -o atiedit -I.. atiedit.c -lgatos ../intl/libintl.a -l
*/
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include "gatos.h"
#include "version.h"
#include "i2c.h"
#include "tvout.h"
#include "m64regs.h"

#define MPPNORMAL       0x80038CB0      /* write */
#define MPPREAD         0x84038CB0      /* prefetch read */
int mppread(int) ;
void mppwrite(int,int) ;
void mppwait(void) ;

int main(int argc, char **argv) {
  char buf[82],*ele;
  u8 datl[32];
  int i, verb=0;
  int addr,n;
  u32 data;
  for ( i=1; i<argc; i++ ) {
    if(!strcmp(argv[i],"-q"))    verb |= GATOSQUIET;
    else if (!strcmp(argv[i],"-v"))    verb |= 1;
    else if (!strcmp(argv[i],"-qv"))   verb |= 1|GATOSQUIET;
    else if (!strcmp(argv[i],"-vq"))   verb |= 1|GATOSQUIET;
    else if (!strncmp(argv[i],"-d",2)) verb |= atoi(argv[i]+2);
    }
  gatos_setverbosity(verb);

  if(!(verb&GATOSQUIET)) {
    printf("\nThis programs is to read/write directly to MEM, MPP, Impact, i2c\n");
    printf("Use as follow: \n");
    printf("   read  mem_0_0a:                0 0x0a \n");
    printf("   write 0x01a0fafb to mem_0_0a:  0 0x0a 01a0fafb\n");
    printf("   read  mem_1_0a:                1 0x0a \n");
    printf("   write 0x01a0fafb to mem_1_0a:  1 0x0a 01a0fafb\n");
    printf("   read  MPP_0xa:                 m 0x0a \n");
    printf("   write 0xf3 to MPP_0a:          m 0x0a f3\n");
    printf("   read  n I2C_0xa:               i n 0x0a \n");
    printf("   write 0xf3 ... to I2C_0a:      i 0 0x0a f3 ...\n");
    printf("   read  tvout_0a:                t 0x0a \n");
    printf("   write 0x01a0fafb to tvout_0a:  t 0x0a 01a0fafb\n");
    printf("   read  tvout_fifo_0a:                f 0x0a \n");
    printf("   write 0x01a0fafb to tvout_fifo_0a:  f 0x0a 01a0fafb\n");
    printf(" NOTE: the data is always an hexadecimal number.\n");
    printf("       The other numbers are written has nnn (decimal)m\n");
    printf("       0ooo (octal) or 0xhh (hexadecimal).\n");
    printf("\n");
  }

  if(gatos_inita()) { perror("xatitv: gatos_inita()"); exit(1); }

  buf[81]='\0';
  while(fgets(buf,80,stdin)!=NULL) {
    ele=strtok(buf," \t");
    if(ele==NULL) continue;
    if(strcmp(ele,"0")==0) {    /* 0 BEGINS */
      if((ele=strtok(NULL," \t\n"))==NULL) continue;
      addr=strtol(ele,NULL,0);
      if((ele=strtok(NULL," \t\n"))!=NULL) {
        data=strtoul(ele,NULL,16);
        (*(MEM_0+addr))=data;
      }
      printf("MEM_0_%02X %08X\n",addr,*(MEM_0+addr));
    } else if(strcmp(ele,"1")==0) { /* 1 BEGINS */
      if((ele=strtok(NULL," \t\n"))==NULL) continue;
      addr=strtol(ele,NULL,0);
      if((ele=strtok(NULL," \t\n"))!=NULL) {
        data=strtoul(ele,NULL,16);
        (*(MEM_1+addr))=data;
      }
      printf("MEM_1_%02X %08X\n",addr,*(MEM_1+addr));
    } else if(strcmp(ele,"m")==0) { /* MPP BEGINS  */
      if((ele=strtok(NULL," \t\n"))==NULL) continue;
      addr=strtol(ele,NULL,0);
      if((ele=strtok(NULL," \t\n"))!=NULL) {
        data=strtoul(ele,NULL,16);
        mppwrite(addr,data);
      }
      printf("MPP_%02X 0x%02X\n",addr,mppread(addr));
    } else if(strcmp(ele,"i")==0) { /* I2C BEGINS */
      if((ele=strtok(NULL," \t\n"))==NULL) continue;
      n=strtol(ele,NULL,0);
      if((ele=strtok(NULL," \t\n"))==NULL) continue;
      addr=strtol(ele,NULL,0);
      i=0;
      while((ele=strtok(NULL," \t\n"))!=NULL)
        datl[i++]=strtoul(ele,NULL,16);
      if(i>0)
        if(i2c_write(addr,datl,i)) 
          printf("Last write not acknowledge!\n");
      if(n>0) {
        if(n>32) n=32;
        i2c_readn(addr,datl,n);
        printf("I2C_%02X",addr);
        for(i=0;i<n;i++)  printf(" 0x%02X",datl[i]);
        printf("\n");
      }
      if(!i2c_device(addr)) printf("Address %02X does not respond!\n", addr);
    } else if(strcmp(ele,"t")==0) { /* TVout BEGINS */
      if((ele=strtok(NULL," \t\n"))==NULL) continue;
      addr=strtol(ele,NULL,0);
      if((ele=strtok(NULL," \t\n"))!=NULL) {
        data=strtoul(ele,NULL,16);
        tvout_write32(addr,data);
      }
      printf("TVO_%02X 0x%08X\n",addr,tvout_read32(addr));
    } else if(strcmp(ele,"f")==0) { /* TVout BEGINS */
      if((ele=strtok(NULL," \t\n"))==NULL) continue;
      addr=strtol(ele,NULL,0);
      if((ele=strtok(NULL," \t\n"))!=NULL) {
        data=strtoul(ele,NULL,16);
        tvout_host_fifo_write(addr,data);
      }
      printf("TVO_FIFO_%02X 0x%08X\n",addr,tvout_host_fifo_read(addr));
    }
    /* flush stdout buffer */
    fflush(stdout);
  }
  return 0;
}

int mppread(int addr) {
  int ret;
  MPP_CONFIG = MPPREAD;
  mppwait();
  MPP_ADDR = addr;
  mppwait();
  ret=MPP_DATA0;
//  tvout_read32(0xa) ;
  return ret;
}

void mppwrite(int addr,int data) {
  mppwait();
  MPP_CONFIG = MPPNORMAL;
  mppwait();
  MPP_ADDR = addr;
  mppwait();
  MPP_DATA0 = data;
//  tvout_read32(0xa) ;
}

void mppwait(void) {
  u32 tries=100;
    while (tries && (MPP_CONFIG3 & 0x40)) tries-- ; if (tries) return ;
  if (!(gatos.debug&GATOSQUIET))
    fprintf(stderr,"mpp_wait(): MPP bus timeout in MPPWAIT\n") ;
}
