/**************************************************************************\
 gatos (General ATI TV and Overlay Software)

  Project Coordinated By Insomnia (Steaphan Greene)
  (insomnia@core.binghamton.edu)

  Copyright (C) 1999 Steaphan Greene, yvind Aabling, Octavian Purdila, 
	Vladimir Dergachev and Christian Lupien.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

\**************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <ibtk/version.h>
#include <ibtk/iwindow.h>
#include <ibtk/ibox.h>
#include <ibtk/itextbox.h>
#include <ibtk/istatbar.h>
#include <ibtk/ibutton.h>
#include <ibtk/isbutton.h>
#include <ibtk/islider.h>
#include <ibtk/iintbox.h>
#include <ibtk/irealbox.h>
#include <ibtk/ilistbox.h>

#include "config.h"
#include "gatos.h"
#include "freq.h"
#include "cc.h"
#include "xutils.h"
#include "version.h"

static IWindow *mwin = NULL;

static int ixs=-1, iys=-1;

void ResizeCB(IWindow *w, int x, int y);
void RepositionCB(IWindow *w, int x, int y);

int main(int argc, char **argv) {
  int i, verb = 0, chkdisp = 1;

  bindtextdomain(PACKAGE, GATOS_LOCALEDIR);
  textdomain(PACKAGE);

  for ( i=1; i<argc; i++ ) {
    if (!strcmp(argv[i],"-nosa")) gatos_setnosa(1) ; /* For testing /AA */
    else if (!strcmp(argv[i],"-pci") && i+1<argc) {
      i++ ; gatos_setpci(argv[i]); /* Override pciaddr in gatos.conf */
      }
    else if (!strcmp(argv[i],"-q"))              verb |= GATOSQUIET;
    else if (!strcmp(argv[i],"-v") 
            || !strcmp(argv[i],"--verbose"))     verb |= 1;
    else if (!strcmp(argv[i],"-qv"))             verb |= 1|GATOSQUIET;
    else if (!strcmp(argv[i],"-vq"))             verb |= 1|GATOSQUIET;
    else if (!strncmp(argv[i],"-d",2))           verb |= atoi(argv[i]+2);
    else if (!strcmp(argv[i],"--force-display")) chkdisp = 0;
    }

  mwin = new IWindow("XATITV-GATOS", "#102030");

  gatos_setverbosity(verb);
  if(VERBOSE) fprintf(stderr, _("IBTK version %s\n"), IBTK_FULL_VERSION_STRING);
  if(gatos_init(GATOS_OVERLAY))
        { perror(_("xatitv: gatos_init()")); exit(1); }
  ixs=gatos_xcapt(); iys=gatos_ycapt();

//  mwin->SetBufSize(640, 480); // This IS correct, even for PAL  --ISG
  mwin->Map(ixs, iys);

/*
  mwin->SetUnhandledCallback(OtherEventCB);
  mwin->SetClickCallback(ClickCB);
  mwin->SetDragCallback(DragCB);
  mwin->SetReleaseCallback(ReleaseCB);
*/
  mwin->SetResizeCallback(ResizeCB);
  mwin->SetRepositionCallback(RepositionCB);
/*
  mwin->SetExposeCallback(ExposeCB);
  mwin->SetActivityCallback(ActivityCB);
  mwin->SetKeyMappingCallback(KeyMappingCB);
*/

  if(gatos_setcolorkey(mwin->GetWinBGColor()))
    { perror(_("xatitv: gatos_setcolorkey()")); exit(1); }
  gatos_enable_sound(1);
  gatos_enable_video(1);
//  gatos_enable_capture(1);

  mwin->DispatchEvents();
  gatos_terminate();
  }

void ResizeCB(IWindow *w, int x, int y) {
  gatos_setgeometry(w->XSize(), w->YSize(), w->XPos(), w->YPos());
  }

void RepositionCB(IWindow *w, int x, int y) {
  gatos_setgeometry(w->XSize(), w->YSize(), w->XPos(), w->YPos());
  }
