/***************************************************************************
                          objects.h  -  description
                             -------------------
    begin                : Tue Dec 7 1999
    copyright            : (C) 1999 by Jeff Anderson
    email                : jbanders@mines.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef MPEG2ENC_H
#	define MPEG2ENC_H
#include "mpeg2enc.h"
#endif

#ifndef STATS_H
# define STATS_H
#include "stats.h"
#endif

#ifndef EXTENSIONS_H
#define EXTENSIONS_H
#include "extensions.h"
#endif

#include <fstream>
class BLOCK
{
	public:
	void set(const BYTE_t colour_comp,const blk_t value);
	void code(void);
	// scan_order == 1 for reqular scan
	// scan_order == 0 for alternate_scan
	void scan(const bool scan_order);
	void write_block(short *predictor);
  ~BLOCK();

	private:
	blk_t block_data;
	int coded_data[8][8];
	int scan_data[64];

	BYTE_t cc;

};


class MACROBLOCK
{
	public:
  MACROBLOCK();
  ~MACROBLOCK();
	void set(const bool first_mb_in_slice, const u16 row, const u16 width,
					 const u16 address,const blk_t Y,const blk_t Cr,const blk_t Cb);
	void code(void);

	void write_macroblock(short *dct_predictor);

	private:
	u16 macroblock_address;

	int macroblock_escape;							// 11-bit
	u16 macroblock_address_increment;		// 1-11 bit vlc
	BYTE_t motion_vectors;
	int macroblock_type;								// 1-9 bit vlc
	BYTE_t quantiser_scale_code;				// 5-bit
	static const BYTE_t marker_bit = 1;

	blk_t *next_Y_block;
	BLOCK *Y_block;
	BLOCK *Cr_block;
	BLOCK *Cb_block;
};



class SLICE
{
 	public:
  SLICE();
	~SLICE();
	void set_slice_start_code(BYTE_t start_code);
	void set_quantiser_scale_code(BYTE_t code);

	void set(const u16 slice_row,const blk_t Y,const blk_t Cr,const blk_t Cb);
	void code(void);
	void write_slice(void);

	private:
	BYTE_t slice_start_code;
	BYTE_t slice_vertical_position_extension;  // 3-bits used if vertical_size > 2800

		// if sequence_SCALABLE_EXTENSION IS PRESENT
		BYTE_t priority_breakpoint; // 7-bit int

	BYTE_t quantiser_scale_code; // 5-bit range: 1-31

		// may not be used
		//if nextbits == 1
		BYTE_t intra_slice_flag; // 1-bit set to 1 for intra_slice and reserved_bits
		BYTE_t intra_slice; //  0 if and macroblocks int the slice are non-intra macroblocks
												// 1 if all macroblocks are intra macroblocks
												// not used to decode
		BYTE_t reserved_bits; // 7-bit int value = 0
	
	BYTE_t extra_bit_slice;  // 1-bit 1 if extra_information_slice will follow
	short extra_information_slice;  // Reserved


	// length is the length of the slice in units of macroblocks
	u32 length;

	u16 row;
	u16 first_macroblock_position;
	short dct_predictor[3];

	MACROBLOCK *mblock;
};

class PICTURE
{
	public:
	PICTURE(BYTE_t picture_type, u16 temporal_reference_value, STATS &mpstat);
	~PICTURE();
	void read_picture();
	void code_picture();
	void write_picture(void);
		
	private:
	static const BYTE_t picture_start_code = 0x00;		// 8-bit
	u16 temporal_reference;														// 10-bit
	BYTE_t picture_coding_type;				// 3-bit
	u16 vbv_delay;										// 16-bit
	static const BYTE_t full_pel_forward_vector = 0;		// 1-bit
	static const BYTE_t forward_f_code = 7;						// 3-bit
	static const BYTE_t full_pel_backward_vector = 0;	// 1-bit
	static const BYTE_t backward_f_code = 7;						// 3-bit
	static const BYTE_t extra_bit_picture = 0;					// 1-bit
// not used but may need changed
//	BYTE_t extra_information_picture;	// 8-bit

	//used internally
	u16 horizontal_size;
	u16 vertical_size;
	u16 color_horizontal_size;
	u16 color_vertical_size;
	u16 slice_count;


	blk_t Y_frame;
	blk_t Cr_frame;
	blk_t Cb_frame;
	blk_t *y_slice, *cr_slice, *cb_slice;
	PICTURE_CODING_EXT pic_code_ext;
	SLICE *slice;
};

class GOP
{
 public:
 GOP(STATS &mpstat);
	void write(void);

 private:
 static const BYTE_t group_start_code = 0xb8;		// 8-bit
 // 25-bit total for next
 BYTE_t drop_frame_flag;									// 1-bit
	// drop_frame_flag is set to 1 only if the frame rate is 29.97
 BYTE_t time_code_hours;									// 5-bits range 0-23
 BYTE_t time_code_minutes;								// 6-bits range 0-59
 static const BYTE_t marker_bit = 1;			// 1-bit value = 1
 BYTE_t time_code_seconds;								// 6-bits range 0-59
 BYTE_t time_code_pictures;								// 6-bits range 0-59

 static const BYTE_t closed_gop = 0;			// 1-bit ; used for editing
 static const BYTE_t broken_link = 0;			// 1-bit value = 0

	PICTURE *pic;
};

class SEQUENCE
{
 public:
	SEQUENCE();
	~SEQUENCE();
	void set_hroz_sz(int size);
	void write(void);
 private:
 	static const BYTE_t sequence_header_code = 0xb3;	// 8-bits
	unsigned int horizontal_size_value;			// 12-bits
	unsigned int vertical_size_value;				// 12-bits
	BYTE_t aspect_ratio_information;				// 4-bits
	unsigned int frame_rate_code;						// 4-bits
// need to figure out how to set the bit_rate_value and vbv_buffer_size to the correct values
	short int bit_rate_value;								// 18-bits
	static const BYTE_t marker_bit = 1;			// 1-bit
	short int vbv_buffer_size_value;				// 10-bits
	static const BYTE_t constrained_param_flag = 0;					// 1-bit
	BYTE_t load_intra_quantiser_matrix;							// 1-bit
	BYTE_t *intra_quantiser_matrix;									// 8-bits x 64
	BYTE_t load_non_intra_quantiser_matrix;					// 1-bit
	BYTE_t *non_intra_quantiser_matrix;							// 8-bits x 64

	GOP *gop;
	SEQUENCE_EXTENSION seq_ext;
};















