/***************************************************************************
                          stats.h  -  description
                             -------------------
    begin                : Sat Dec 18 1999
    copyright            : (C) 1999 by Jeff Anderson
    email                : jbanders@mines.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MPEG2ENC_H
#define MPEG2ENC_H
#include "mpeg2enc.h"
#endif

class STATS
{
	public:
	STATS();

	//return 12 least significant bits of horizontal_size
	u16 get_horizontal_size_value(void);

	u32 get_horizontal_size(void);
	u32 get_color_horizontal_size(void);

	//return 12 least significant bits of vertical_size
	u16 get_vertical_size_value(void);

	u32 get_vertical_size(void);
	u32 get_color_vertical_size(void);

	BYTE_t get_aspect_ratio_information(void);
	BYTE_t get_frame_rate_code(void);
	BYTE_t get_load_intra_quantiser_matrix(void);
	BYTE_t get_load_non_intra_quantiser_matrix(void);

	//used by gop
	BYTE_t get_drop_frame_flag(void);
	BYTE_t get_time_code_hours(void);
	BYTE_t get_time_code_minutes(void);
	BYTE_t get_time_code_seconds(void);
	BYTE_t get_time_code_pictures(void);

	u16 get_temporal_reference(void);
	u16 get_vbv_delay(void);
	u16 get_numb_slice(void);

	u32 get_slice_length(void);

	bool get_scan_type(void);

	// set functions
	void set_vertical_size(int vert_size);
	void set_horizontal_size(int horz_size);

	void set_frame_rate(char *frame_rate);
	private:
	u32 horizontal_size;
	u32 vertical_size;

	//from sequence
	BYTE_t aspect_ratio;
/*
	aspect_ratio_information table

	aspect_ratio_information	Sample Aspect		DAR
														Ratio
	0000											Forbidden				Forbidden
	0001											1.0 Square			-
														sample
	0010											-								3/4
	0011											-								9/16
	0100											-								1/2.21
	0101											-								Reserved
	...																				...
	1111											-								Reserved
*/

	BYTE_t frame_rate_code;
/*
	frame_rate_value

	frame_rate_code		frame_rate_value
	0000							Forbidden
	0001							24/1.001 (23.976...)
	0010							24
	0011							25
	0100							30.000/1.001 (29.97...)
	0101							30
	0110							50
	0111							60.000/1.001 (59.94...)
	1000							60
	1001							Reserved
	...								...
	1111							Reserved
*/
	// next may need fixing
	BYTE_t temporal_picture_stucture;
	BYTE_t chrominance_format;
	u32 bit_rate;
	u32 vbv_buffer_size;
	BYTE_t *intra_quantiser_matrix;
	BYTE_t *non_intra_quantiser_matrix;

	//used by GOP
 // 25-bit total for next
	BYTE_t drop_frame_flag;									// 1-bit
	BYTE_t time_code_hours;									// 5-bits range 0-23
	BYTE_t time_code_minutes;								// 6-bits range 0-59
	BYTE_t time_code_seconds;								// 6-bits range 0-59
	BYTE_t time_code_pictures;								// 6-bits range 0-59

	//used by PICTURE header
	u16 temporal_reference;						// 10-bit
	BYTE_t picture_coding_type;				// 3-bit
	u16 vbv_delay;										// 16-bit

	//determines scan type
	bool alternate_scan;	
};