/**
 *
 * Copyright (C) 2000 Vandoorselaere Yoann, Thomas Mirlacher
 * All Rights Reserved
 *
 * This file is part of the OMS program.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "xmlparse.h"
#include "xmltag_widget.h"

static oms_xmltag_tbl_t xmltag_tbl_widget[] = {
        { "checkbox",   CHECKBOX	},
        { "radio",      RADIO		},
        { "textfield",  TEXTFIELD	},
        { "input",  	INPUT		},
        { NULL,         NOTYPE		},
};


static void handle_tag_input (xmlDocPtr doc, xmlNodePtr cur);

static xmlhandler_tbl_t xmlhandler_tbl[] = {
	{"input", handle_tag_input},
	{"checkbox", handle_tag_input},
	{NULL, NULL},
};


xmlhandler_tbl_t *xmltag_widget_init (void)
{
	return xmlhandler_tbl;
}


/**
 *
 **/

static void handle_tag_input (xmlDocPtr doc, xmlNodePtr cur) 
{
        char *name;
        char *type_name;
	int i;
        uint type = NOTYPE;

// parse name and type, which HAVE to be present

	if (!(name = xmlGetProp (cur, "name"))) {
		fprintf (stderr, "xmlGetProp(\"name\") returned NULL\n");
		return;
	}
	if (!(type_name = xmlGetProp (cur, "type"))) {
		fprintf (stderr, "xmlGetProp(\"type\") returned NULL\n");
		return;
	}

//fprintf (stderr, "%s/%s\n", name, type_name);

	for (i=0; xmltag_tbl_widget[i].name; i++) {
		if (!strcmp (type_name, xmltag_tbl_widget[i].name)) {
			type = xmltag_tbl_widget[i].type;
			break;
		}        
	}

// Create a node, add it to the list and fill it.
	if (name && (type != NOTYPE)) {
        	oms_xmltree_t *item;
		oms_xmltree_node_t *data;
	
		if (!(data = (void *) calloc (1, sizeof (oms_xmltree_node_t)))) {
			perror ("calloc");
			return;
		}

		data->type_name	= type_name;
		data->type_val	= type;
		data->var_name	= name;
		data->var_val	= xmlGetProp (cur, "value");
		data->text	= xmlNodeListGetString (doc, cur->childs, 1);

		if (!(item = oms_xml_list_add (data))) {
			fprintf (stderr, "error adding new node");
               		return;
		}
	}
}
