/*****
*
* This file is part of the OMS program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <sys/time.h>
#include <inttypes.h>
#include <unistd.h>
#include "scr.h"

static struct timeval initial_time;
static uint32_t initial_scr;

void set_scr (uint32_t new_scr)
{
	gettimeofday (&initial_time, NULL);
	initial_scr = new_scr;
}

uint32_t get_scr (void)
{
	struct timeval current_time;
	unsigned int difference;
	uint32_t current_scr;

	gettimeofday (&current, NULL);

	// stored in terms of usec
	difference = (current_time.tv_usec - initial_time.tv_usec) +
	    ((current_time.tv_sec - initial_time.tv_sec) * 1000000);
	current_scr = initial_scr + (difference / 100 * 9);

	// returns SCR in terms of its 90kHz clock
	return current_scr;
}

int wait_for_scr (uint32_t final_scr)
{
	uint32_t current_scr = get_scr ();

	// FIXME: OVERFLOW
	if (current_scr > final_scr)	// we can't wait for the past...
		return 1;

	usleep ((unsigned int) ((current_scr - final_scr) / 9 * 100));	// FIXME: check for error

	return 0;
}
