
/*
 *
 * Copyright (C) 1999-2000	Thomas Mirlacher,
 *	         1999		Mats Sjvall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The authors may be reached as <dent@linuxvideo.org>, or
 *	<matsj398@student.liu.se>
 *
 *------------------------------------------------------------
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>

#include <oms/oms.h>
#include <oms/plugin/decaps.h>

static int _vcd_open	(void *self, void *name);
static int _vcd_close	(void *self);
static int _vcd_decaps	(buf_t *buf);
static int _vcd_ctrl	(void *self, uint ctrl_id, ...);

typedef struct {
	uint8_t sync		[12];
	uint8_t header		[4];
	uint8_t subheader	[8];
	uint8_t data		[2324];
	uint8_t spare		[4];
} cdsector_t;

static plugin_decaps_magic_t _vcd_magic = {
	NULL,
	0,
	NULL
};

static plugin_decaps_t decaps = {
	open:		_vcd_open,
	close:		_vcd_close,
	decaps:		_vcd_decaps,
	ctrl:		_vcd_ctrl,
	magic:		&_vcd_magic,
	blocksize:	sizeof(cdsector_t),
        config:         NULL,
};

static int fd_out;

static int _vcd_open (void *self, void *name)
{
	fd_out = open ("/tmp/vcd.out", O_WRONLY);
	return 0;
}


static int _vcd_close (void *self)
{
	close (fd_out);
	return 0;
}


static int _vcd_decaps (buf_t *_buf)
{
	static buf_entry_t *buf_entry = NULL;
	cdsector_t *cd_buf;
	buf_t *buf = _buf;

	if (!(buf_entry = oms_buf_alloc (buf, sizeof (cdsector_t))))
		return 0;

	cd_buf = (cdsector_t *) buf_entry->mem;

	if ((buf_entry->mem_len = decaps.read (NULL, buf_entry->mem, decaps.blocksize)) <= 0)
		return -1;
  
	if (cd_buf->subheader[1] == 0x01) {
		if ((cd_buf->subheader[2] == 0x62) && (cd_buf->subheader[3] == 0x0f))
			buf_entry->buf_id = BUF_VIDEO;
		else if ((cd_buf->subheader[2] == 0x64) && (cd_buf->subheader[3] == 0x7f))
			buf_entry->buf_id = BUF_AUDIO;
		else {
			LOG (LOG_WARNING, "Unhandled stream: %x%x\n",
				cd_buf->subheader[2],
				cd_buf->subheader[3]);
			return -1;
		}


		buf_entry->data		= cd_buf->data;
		buf_entry->data_len	= (size_t) sizeof (cd_buf->data);

                write (fd_out, buf_entry->data, buf_entry->data_len);
//		decaps.write (buf, buf_entry);
	fprintf (stderr, "+");
	}
	//fprintf (stderr, "-");

	return 0;
}


static int _vcd_ctrl (void *self, uint ctrl_id, ...)
{
	return 0;
}


int PLUGIN_INIT (decaps_vcd) (char *whoami)
{
	pluginRegister (whoami,
		PLUGIN_ID_DECAPS,
		"vcd",
		NULL,
		NULL,
		&decaps);

	return 0;
}


void PLUGIN_EXIT (decaps_vcd) (void)
{
	LOG (LOG_DEBUG, "decaps_close called\n");
}
