
/*
 *
 * Copyright (C) 1999, 2000  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 */


//#define TOAST_SPU
//#define TOAST_SYNC 1

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>

#include <oms/oms.h>
#include <oms/plugin.h>

#ifdef TOAST_SPU
#include "../../codec/spudec/spu.h"
#include "../../codec/spudec/overlay.h"
#include "../../codec/spudec/display.h"

static u_int *bg_image;
#endif

#include <ac3.h>
#include <mpeg2.h>

// FIXME for the fps display - should be removed
#include <sys/time.h>

static int _sw_init (plugin_codec_t *plugin, buf_t *buf, char *output_path);
static int _sw_open (plugin_codec_t *plugin);
static int _sw_read (plugin_codec_t *plugin, buf_t *buf, buf_entry_t *buf_entry);
static int _sw_set_status (plugin_codec_t *plugin, u_int status);
static int _sw_set_attributes (plugin_codec_t *plugin, u_int id, void *attr);

static struct sw_priv_struct {
	int fd_video;
	int fd_audio;
	int fd_other;
#ifndef TOAST_SPU
	int fd_spu;
#endif
	vo_functions_t *video_out;
	buf_t *buf;
} sw_priv;


static plugin_codec_t codec_sw = {
	&sw_priv,
	_sw_init,	// init
	NULL,		// exit
	_sw_open,	// open
	NULL,		// close
	_sw_read,	// read
	NULL,		// write
	_sw_set_status,
	_sw_set_attributes,
}; 


static ac3_config_t ac3_config;


/**
 *
 **/

static int _sw_open (plugin_codec_t *plugin)
{
	return 0;
}

/*
 * stolen from mpeg2dec
 * FIXME this -really- should be somewhere else... like the UI code
 */
static struct timeval tv_beg, tv_end, tv_start;
static unsigned int frame_counter = 0;
static unsigned int elapsed;
static unsigned int total_elapsed;
static unsigned int last_count = 0;
static void print_fps(uint_32 final) 
{
	int fps, tfps, frames;
	
	gettimeofday(&tv_end, NULL);
	if (frame_counter++ == 0) {
		tv_start = tv_beg = tv_end;
	}
	elapsed = (tv_end.tv_sec - tv_beg.tv_sec) * 100 +
		  (tv_end.tv_usec - tv_beg.tv_usec) / 10000;

	total_elapsed = (tv_end.tv_sec - tv_start.tv_sec) * 100 +
		  (tv_end.tv_usec - tv_start.tv_usec) / 10000;

	if (final) {
		tfps = frame_counter * 10000 / total_elapsed;
		fprintf(stderr,"\n%d frames decoded in %d.%02d "
			"seconds (%d.%02d fps)\n", frame_counter,
			total_elapsed / 100, total_elapsed % 100,
			tfps / 100, tfps % 100);
		return;
	}

	if (elapsed < 50)	/* only display every 0.50 seconds */
		return;

	tv_beg = tv_end;
	frames = frame_counter - last_count;

	fps = frames * 10000 / elapsed;			/* 100x */
	tfps = frame_counter * 10000 / total_elapsed;	/* 100x */

	fprintf(stderr, "%d frames in %d.%02d sec (%d.%02d fps), "
		"%d last %d.%02d sec (%d.%02d fps)\033[K\r", frame_counter,
		total_elapsed / 100, total_elapsed % 100,
		tfps / 100, tfps % 100, frames, elapsed / 100, elapsed % 100,
		fps / 100, fps % 100);

	last_count = frame_counter;
}


/**
 *
 **/

static int _sw_read (plugin_codec_t *plugin, buf_t *buf, buf_entry_t *buf_entry)
{
	struct sw_priv_struct *priv = plugin->priv; 
	int i;
	unsigned int num_frames;
  
#if 0
	if (buf_entry->flags & BUF_FLAG_CLOCK_VALID) {
		u_int64_t clock = (buf_entry->clock * 300 + buf_entry->clock_ext)/27; // MHz
		bufSetTime (buf, clock);
	}
#endif

	switch (buf_entry->buf_id) {
		case BUF_AUDIO:
//			switch (buf_entry->type) 
//				default:
			ac3_decode_data (buf_entry->data, buf_entry->data+buf_entry->data_len);  

		break;

		case BUF_OTHER:
			if (priv->fd_other >= 0) {
				fprintf (stderr, "other %d",  buf_entry->data_len);
				write (priv->fd_other, buf_entry->data, buf_entry->data_len);
			}
    
		break;

		case BUF_SUBPIC: {
#ifndef TOAST_SPU
			if (priv->fd_spu >= 0)
				write (priv->fd_spu, buf_entry->data, buf_entry->data_len);
#else
			struct spu_img_struct spu;
    
			if (!spuParseHdr (&spu, buf_entry->data-3, buf_entry->data_len+3)) {
				spuParseData (&spu);
				memset (bg_image, 0, 720 * 576 * sizeof(u_int));
				if ((bg_image = overlay (bg_image, &spu, 0.8)))
					display_frame (bg_image);
					free (spu.data);
				}
#endif
		}
		break;

		case BUF_VIDEO:
			num_frames = mpeg2_decode_data (priv->video_out, buf_entry->data, buf_entry->data+buf_entry->data_len);  
			while (num_frames--)
				print_fps(0);

		break;

		default:
#ifdef DEBUG
			fprintf (stderr, "codec_sw: unhandled stream type\n");
#endif
	}

	return 0;
}


/**
 *
 **/

static int _sw_init (plugin_codec_t *plugin, buf_t *_buf, char *output_path)
{
	int i;
	vo_functions_t *vo_funcs = NULL;
	ao_functions_t *ao_funcs = NULL;
	//FIXME hack output_path format "videoname:audioname"
	char buf[100], *vo, *ao;

	struct sw_priv_struct *priv = plugin->priv; 

	priv->buf = _buf;

	// FIXME
	strncpy(buf,output_path,100);
	vo = strtok(buf, ":, ");
	ao = strtok(NULL, ":, ");
	fprintf(stderr,"v:%s a:%s\n", vo, ao);

	for (i=0; video_out_drivers[i] != NULL && vo_funcs == NULL; i++) {
		const vo_info_t* info = video_out_drivers[i]->get_info();
		if (strcmp(vo,info->short_name) == 0) {
			vo_funcs = video_out_drivers[i];
		}
	}
	if (vo_funcs == NULL) {
		vo_funcs = video_out_drivers[0];
	}
	priv->video_out = vo_funcs;
	mpeg2_init ();

        ac3_config.num_output_ch = 2;
        ac3_config.flags = 0;
     
	for (i=0; audio_out_drivers[i] != NULL && ao_funcs == NULL; i++) {
		const ao_info_t* info = audio_out_drivers[i]->get_info();
		if (strcmp(ao,info->short_name) == 0) {
			ao_funcs = audio_out_drivers[i];
		}
	}
	if (ao_funcs == NULL) {
		ao_funcs = audio_out_drivers[0];
	}
	ac3_init (&ac3_config, ao_funcs);
   
	priv->fd_audio = 0;
	fprintf (stderr, "Trying to open /tmp/other... ");
	if ((priv->fd_other = open ("/tmp/other", O_WRONLY)) < 0)
		fprintf (stderr, "failed: other streams will be ignored.\n");
	else
		fprintf (stderr, "success.\n");

#ifndef TOAST_SPU
	fprintf (stderr, "Trying to open /tmp/spu... ");
	if ((priv->fd_spu = open ("/tmp/spu", O_WRONLY)) < 0)
		fprintf (stderr, "failed: spu stream will be ignored.\n");
	else
		fprintf (stderr, "success.\n");
#else
	spuInit ();

	bg_image = (u_int *) calloc (1, 720 * 576 * sizeof(u_int));
#endif

	return 0;
}


/**
 *
 **/

static int _sw_set_status (plugin_codec_t *plugin, u_int status)
{
	switch (status) {
	case STATUS_PLAY:
	case STATUS_PAUSE:
	case STATUS_STOP:
		break;
	default:
		fprintf (stderr, "DecoderSetStatus: unhandled status %d\n", status);
		return -1;
	}

	return 0;
}


/**
 *
 **/

static int _sw_set_attributes (plugin_codec_t *plugin, u_int id, void *attr)
{
#ifdef TOAST_SPU
#ifdef DEBUG
	fprintf (stderr, "setting CLUT\n");
#endif
	switch (id) {
	case BUF_SUBPIC: {	
		plugin_codec_attr_spu_t *subpic = attr;
		spuSetCLUT (subpic->clut);
		break;
	}	
	}
#endif
	return 0;
}


/**
 * Initialize Plugin.
 **/

void *plugin_init (char *whoami)
{
	pluginRegister (whoami,
		PLUGIN_ID_CODEC_VIDEO,
		STREAM_ID_MPEG2,
		&codec_sw);

	return &codec_sw;
}


/**
 * Cleanup Plugin.
 **/

void plugin_exit (void)
{
}
